from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional, Union

from flet.core.alignment import Alignment
from flet.core.border import BorderSide
from flet.core.border_radius import BorderRadius
from flet.core.box import BoxConstraints, BoxDecoration, BoxShadow
from flet.core.buttons import ButtonStyle, OutlinedBorder
from flet.core.menu_bar import MenuStyle
from flet.core.navigation_bar import NavigationBarLabelBehavior
from flet.core.navigation_rail import NavigationRailLabelType
from flet.core.popup_menu_button import PopupMenuPosition
from flet.core.size import Size
from flet.core.slider import SliderInteraction
from flet.core.snack_bar import DismissDirection, SnackBarBehavior
from flet.core.text_style import TextStyle
from flet.core.textfield import TextCapitalization
from flet.core.tooltip import TooltipTriggerMode
from flet.core.types import (
    Brightness,
    ClipBehavior,
    ColorValue,
    ControlState,
    ControlStateValue,
    DurationValue,
    IconValue,
    Locale,
    MainAxisAlignment,
    MarginValue,
    MouseCursor,
    NotchShape,
    OffsetValue,
    OptionalNumber,
    PaddingValue,
    StrokeCap,
    TextAlign,
    VisualDensity,
)
from flet.utils.deprecated import deprecated_class, deprecated_property

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class PageTransitionTheme(Enum):
    NONE = "none"
    FADE_UPWARDS = "fadeUpwards"
    OPEN_UPWARDS = "openUpwards"
    ZOOM = "zoom"
    CUPERTINO = "cupertino"
    PREDICTIVE = "predictive"
    FADE_FORWARDS = "fadeForwards"


@dataclass
class PageTransitionsTheme:
    android: Optional[PageTransitionTheme] = None
    ios: Optional[PageTransitionTheme] = None
    linux: Optional[PageTransitionTheme] = None
    macos: Optional[PageTransitionTheme] = None
    windows: Optional[PageTransitionTheme] = None


@dataclass
class ColorScheme:
    primary: Optional[ColorValue] = None
    on_primary: Optional[ColorValue] = None
    primary_container: Optional[ColorValue] = None
    on_primary_container: Optional[ColorValue] = None
    secondary: Optional[ColorValue] = None
    on_secondary: Optional[ColorValue] = None
    secondary_container: Optional[ColorValue] = None
    on_secondary_container: Optional[ColorValue] = None
    tertiary: Optional[ColorValue] = None
    on_tertiary: Optional[ColorValue] = None
    tertiary_container: Optional[ColorValue] = None
    on_tertiary_container: Optional[ColorValue] = None
    error: Optional[ColorValue] = None
    on_error: Optional[ColorValue] = None
    error_container: Optional[ColorValue] = None
    on_error_container: Optional[ColorValue] = None
    background: Optional[ColorValue] = None
    on_background: Optional[ColorValue] = None
    surface: Optional[ColorValue] = None
    on_surface: Optional[ColorValue] = None
    surface_variant: Optional[ColorValue] = None
    on_surface_variant: Optional[ColorValue] = None
    outline: Optional[ColorValue] = None
    outline_variant: Optional[ColorValue] = None
    shadow: Optional[ColorValue] = None
    scrim: Optional[ColorValue] = None
    inverse_surface: Optional[ColorValue] = None
    on_inverse_surface: Optional[ColorValue] = None
    inverse_primary: Optional[ColorValue] = None
    surface_tint: Optional[ColorValue] = None
    on_primary_fixed: Optional[ColorValue] = None
    on_secondary_fixed: Optional[ColorValue] = None
    on_tertiary_fixed: Optional[ColorValue] = None
    on_primary_fixed_variant: Optional[ColorValue] = None
    on_secondary_fixed_variant: Optional[ColorValue] = None
    on_tertiary_fixed_variant: Optional[ColorValue] = None
    primary_fixed: Optional[ColorValue] = None
    secondary_fixed: Optional[ColorValue] = None
    tertiary_fixed: Optional[ColorValue] = None
    primary_fixed_dim: Optional[ColorValue] = None
    secondary_fixed_dim: Optional[ColorValue] = None
    surface_bright: Optional[ColorValue] = None
    surface_container: Optional[ColorValue] = None
    surface_container_high: Optional[ColorValue] = None
    surface_container_low: Optional[ColorValue] = None
    surface_container_lowest: Optional[ColorValue] = None
    surface_dim: Optional[ColorValue] = None
    tertiary_fixed_dim: Optional[ColorValue] = None


@dataclass
class TextTheme:
    body_large: Optional[TextStyle] = None
    body_medium: Optional[TextStyle] = None
    body_small: Optional[TextStyle] = None
    display_large: Optional[TextStyle] = None
    display_medium: Optional[TextStyle] = None
    display_small: Optional[TextStyle] = None
    headline_large: Optional[TextStyle] = None
    headline_medium: Optional[TextStyle] = None
    headline_small: Optional[TextStyle] = None
    label_large: Optional[TextStyle] = None
    label_medium: Optional[TextStyle] = None
    label_small: Optional[TextStyle] = None
    title_large: Optional[TextStyle] = None
    title_medium: Optional[TextStyle] = None
    title_small: Optional[TextStyle] = None


@dataclass
class ScrollbarTheme:
    thumb_visibility: ControlStateValue[bool] = None
    thickness: ControlStateValue[OptionalNumber] = None
    track_visibility: ControlStateValue[bool] = None
    radius: Optional[float] = None
    thumb_color: ControlStateValue[ColorValue] = None
    track_color: ControlStateValue[ColorValue] = None
    track_border_color: ControlStateValue[ColorValue] = None
    cross_axis_margin: Optional[float] = None
    main_axis_margin: Optional[float] = None
    min_thumb_length: Optional[float] = None
    interactive: Optional[bool] = None

    def __post_init__(self):
        if not isinstance(self.thumb_visibility, dict):
            self.thumb_visibility = {ControlState.DEFAULT: self.thumb_visibility}
        if not isinstance(self.thickness, dict):
            self.thickness = {ControlState.DEFAULT: self.thickness}
        if not isinstance(self.track_visibility, dict):
            self.track_visibility = {ControlState.DEFAULT: self.track_visibility}
        if not isinstance(self.thumb_color, dict):
            self.thumb_color = {ControlState.DEFAULT: self.thumb_color}
        if not isinstance(self.track_color, dict):
            self.track_color = {ControlState.DEFAULT: self.track_color}
        if not isinstance(self.track_border_color, dict):
            self.track_border_color = {ControlState.DEFAULT: self.track_border_color}


@dataclass
class TabsTheme:
    divider_color: Optional[ColorValue] = None
    indicator_border_radius: Optional[BorderRadius] = None
    indicator_border_side: Optional[BorderSide] = None
    indicator_padding: Optional[PaddingValue] = None
    indicator_color: Optional[ColorValue] = None
    indicator_tab_size: Optional[bool] = None
    label_color: Optional[ColorValue] = None
    unselected_label_color: Optional[ColorValue] = None
    overlay_color: ControlStateValue[ColorValue] = None
    mouse_cursor: ControlStateValue[MouseCursor] = None
    label_padding: Optional[PaddingValue] = None
    label_text_style: Optional[TextStyle] = None
    unselected_label_text_style: Optional[TextStyle] = None

    def __post_init__(self):
        if not isinstance(self.overlay_color, dict):
            self.overlay_color = {ControlState.DEFAULT: self.overlay_color}
        if not isinstance(self.mouse_cursor, dict):
            self.mouse_cursor = {ControlState.DEFAULT: self.mouse_cursor}


@dataclass
class SystemOverlayStyle:
    status_bar_color: Optional[ColorValue] = None
    system_navigation_bar_color: Optional[ColorValue] = None
    system_navigation_bar_divider_color: Optional[ColorValue] = None
    enforce_system_navigation_bar_contrast: Optional[bool] = None
    enforce_system_status_bar_contrast: Optional[bool] = None
    system_navigation_bar_icon_brightness: Optional[Brightness] = None
    status_bar_brightness: Optional[Brightness] = None
    status_bar_icon_brightness: Optional[Brightness] = None


@dataclass
class DialogTheme:
    bgcolor: Optional[ColorValue] = None
    shadow_color: Optional[ColorValue] = None
    surface_tint_color: Optional[ColorValue] = None
    icon_color: Optional[ColorValue] = None
    elevation: OptionalNumber = None
    shape: Optional[OutlinedBorder] = None
    title_text_style: Optional[TextStyle] = None
    content_text_style: Optional[TextStyle] = None
    alignment: Optional[Alignment] = None
    actions_padding: Optional[PaddingValue] = None
    clip_behavior: Optional[ClipBehavior] = None
    barrier_color: Optional[ColorValue] = None
    inset_padding: Optional[PaddingValue] = None


@dataclass
class ElevatedButtonTheme:
    bgcolor: Optional[ColorValue] = None
    foreground_color: Optional[ColorValue] = None
    icon_color: Optional[ColorValue] = None
    shadow_color: Optional[ColorValue] = None
    disabled_bgcolor: Optional[ColorValue] = None
    disabled_foreground_color: Optional[ColorValue] = None
    disabled_icon_color: Optional[ColorValue] = None
    overlay_color: Optional[ColorValue] = None
    surface_tint_color: Optional[ColorValue] = None
    elevation: OptionalNumber = None
    padding: Optional[PaddingValue] = None
    enable_feedback: Optional[bool] = None
    disabled_mouse_cursor: Optional[MouseCursor] = None
    enabled_mouse_cursor: Optional[MouseCursor] = None
    shape: Optional[OutlinedBorder] = None
    text_style: Optional[TextStyle] = None
    visual_density: Optional[VisualDensity] = None
    border_side: Optional[BorderSide] = None
    animation_duration: Optional[DurationValue] = None
    alignment: Optional[Alignment] = None
    icon_size: OptionalNumber = None
    fixed_size: Optional[Size] = None
    maximum_size: Optional[Size] = None
    minimum_size: Optional[Size] = None


@dataclass
class OutlinedButtonTheme(ElevatedButtonTheme):
    pass


@dataclass
class TextButtonTheme(ElevatedButtonTheme):
    pass


@dataclass
class FilledButtonTheme(ElevatedButtonTheme):
    pass


@dataclass
class IconButtonTheme:
    # from ElevatedButtonTheme (excluding icon_color, disabled_icon_color, text_style)
    bgcolor: Optional[ColorValue] = None
    foreground_color: Optional[ColorValue] = None
    shadow_color: Optional[ColorValue] = None
    disabled_bgcolor: Optional[ColorValue] = None
    disabled_foreground_color: Optional[ColorValue] = None
    overlay_color: Optional[ColorValue] = None
    surface_tint_color: Optional[ColorValue] = None
    elevation: OptionalNumber = None
    padding: Optional[PaddingValue] = None
    enable_feedback: Optional[bool] = None
    disabled_mouse_cursor: Optional[MouseCursor] = None
    enabled_mouse_cursor: Optional[MouseCursor] = None
    shape: Optional[OutlinedBorder] = None
    visual_density: Optional[VisualDensity] = None
    border_side: Optional[BorderSide] = None
    animation_duration: Optional[DurationValue] = None
    alignment: Optional[Alignment] = None
    icon_size: OptionalNumber = None
    fixed_size: Optional[Size] = None
    maximum_size: Optional[Size] = None
    minimum_size: Optional[Size] = None
    # Icon Button Theme
    focus_color: Optional[ColorValue] = None
    highlight_color: Optional[ColorValue] = None
    hover_color: Optional[ColorValue] = None


@dataclass
class BottomSheetTheme:
    bgcolor: Optional[ColorValue] = None
    shadow_color: Optional[ColorValue] = None
    surface_tint_color: Optional[ColorValue] = None
    drag_handle_color: Optional[ColorValue] = None
    elevation: OptionalNumber = None
    shape: Optional[OutlinedBorder] = None
    show_drag_handle: Optional[bool] = None
    modal_bgcolor: Optional[ColorValue] = None
    modal_elevation: OptionalNumber = None
    clip_behavior: Optional[ClipBehavior] = None
    size_constraints: Optional[BoxConstraints] = None
    modal_barrier_color: Optional[ColorValue] = None


@dataclass
class CardTheme:
    color: Optional[ColorValue] = None
    shadow_color: Optional[ColorValue] = None
    surface_tint_color: Optional[ColorValue] = None
    elevation: OptionalNumber = None
    shape: Optional[OutlinedBorder] = None
    clip_behavior: Optional[ClipBehavior] = None
    margin: Optional[MarginValue] = None


@dataclass
class ChipTheme:
    color: ControlStateValue[ColorValue] = None
    bgcolor: Optional[ColorValue] = None
    shadow_color: Optional[ColorValue] = None
    surface_tint_color: Optional[ColorValue] = None
    disabled_color: Optional[ColorValue] = None
    selected_color: Optional[ColorValue] = None
    checkmark_color: Optional[ColorValue] = None
    delete_icon_color: Optional[ColorValue] = None
    secondary_selected_color: Optional[ColorValue] = None
    selected_shadow_color: Optional[ColorValue] = None
    elevation: OptionalNumber = None
    click_elevation: OptionalNumber = None
    shape: Optional[OutlinedBorder] = None
    padding: Optional[PaddingValue] = None
    label_padding: Optional[PaddingValue] = None
    label_text_style: Optional[TextStyle] = None
    secondary_label_text_style: Optional[TextStyle] = None
    border_side: Optional[BorderSide] = None
    brightness: Optional[Brightness] = None
    show_checkmark: Optional[bool] = None
    avatar_constraints: Optional[BoxConstraints] = None
    delete_icon_size_constraints: Optional[BoxConstraints] = None

    def __post_init__(self):
        if not isinstance(self.color, dict):
            self.color = {ControlState.DEFAULT: self.color}


@dataclass
class FloatingActionButtonTheme:
    bgcolor: Optional[ColorValue] = None
    hover_color: Optional[ColorValue] = None
    focus_color: Optional[ColorValue] = None
    foreground_color: Optional[ColorValue] = None
    splash_color: Optional[ColorValue] = None
    elevation: OptionalNumber = None
    focus_elevation: OptionalNumber = None
    hover_elevation: OptionalNumber = None
    highlight_elevation: OptionalNumber = None
    disabled_elevation: OptionalNumber = None
    shape: Optional[OutlinedBorder] = None
    enable_feedback: Optional[bool] = None
    extended_padding: Optional[PaddingValue] = None
    extended_text_style: Optional[TextStyle] = None
    extended_icon_label_spacing: OptionalNumber = None
    extended_size_constraints: Optional[BoxConstraints] = None
    size_constraints: Optional[BoxConstraints] = None
    large_size_constraints: Optional[BoxConstraints] = None
    small_size_constraints: Optional[BoxConstraints] = None


@dataclass
class NavigationRailTheme:
    bgcolor: Optional[ColorValue] = None
    indicator_color: Optional[ColorValue] = None
    elevation: OptionalNumber = None
    indicator_shape: Optional[OutlinedBorder] = None
    unselected_label_text_style: Optional[TextStyle] = None
    selected_label_text_style: Optional[TextStyle] = None
    label_type: Optional[NavigationRailLabelType] = None
    min_width: OptionalNumber = None
    min_extended_width: OptionalNumber = None
    group_alignment: OptionalNumber = None
    use_indicator: Optional[bool] = None


@dataclass
class AppBarTheme:
    color: Optional[ColorValue] = None
    bgcolor: Optional[ColorValue] = None
    shadow_color: Optional[ColorValue] = None
    surface_tint_color: Optional[ColorValue] = None
    foreground_color: Optional[ColorValue] = None
    elevation: OptionalNumber = None
    shape: Optional[OutlinedBorder] = None
    title_text_style: Optional[TextStyle] = None
    toolbar_text_style: Optional[TextStyle] = None
    center_title: Optional[bool] = None
    title_spacing: OptionalNumber = None
    scroll_elevation: OptionalNumber = None
    toolbar_height: OptionalNumber = None
    actions_padding: Optional[PaddingValue] = None


@dataclass
class BottomAppBarTheme:
    color: Optional[ColorValue] = None
    shadow_color: Optional[ColorValue] = None
    surface_tint_color: Optional[ColorValue] = None
    elevation: OptionalNumber = None
    height: OptionalNumber = None
    padding: Optional[PaddingValue] = None
    shape: Optional[NotchShape] = None


@dataclass
class RadioTheme:
    fill_color: ControlStateValue[ColorValue] = None
    overlay_color: ControlStateValue[ColorValue] = None
    splash_radius: OptionalNumber = None
    height: OptionalNumber = None
    visual_density: Optional[VisualDensity] = None
    mouse_cursor: ControlStateValue[MouseCursor] = None

    def __post_init__(self):
        if not isinstance(self.fill_color, dict):
            self.fill_color = {ControlState.DEFAULT: self.fill_color}
        if not isinstance(self.overlay_color, dict):
            self.overlay_color = {ControlState.DEFAULT: self.overlay_color}
        if not isinstance(self.mouse_cursor, dict):
            self.mouse_cursor = {ControlState.DEFAULT: self.mouse_cursor}


@dataclass
class CheckboxTheme:
    overlay_color: ControlStateValue[ColorValue] = None
    check_color: ControlStateValue[ColorValue] = None
    fill_color: ControlStateValue[ColorValue] = None
    splash_radius: OptionalNumber = None
    border_side: Optional[BorderSide] = None
    visual_density: Optional[VisualDensity] = None
    shape: Optional[OutlinedBorder] = None
    mouse_cursor: ControlStateValue[MouseCursor] = None

    def __post_init__(self):
        if not isinstance(self.overlay_color, dict):
            self.overlay_color = {ControlState.DEFAULT: self.overlay_color}
        if not isinstance(self.check_color, dict):
            self.check_color = {ControlState.DEFAULT: self.check_color}
        if not isinstance(self.fill_color, dict):
            self.fill_color = {ControlState.DEFAULT: self.fill_color}
        if not isinstance(self.mouse_cursor, dict):
            self.mouse_cursor = {ControlState.DEFAULT: self.mouse_cursor}


@dataclass
class BadgeTheme:
    bgcolor: Optional[ColorValue] = None
    text_color: Optional[ColorValue] = None
    small_size: OptionalNumber = None
    large_size: OptionalNumber = None
    alignment: Optional[Alignment] = None
    padding: Optional[PaddingValue] = None
    offset: Optional[OffsetValue] = None
    text_style: Optional[TextStyle] = None


@dataclass
class SwitchTheme:
    thumb_color: ControlStateValue[ColorValue] = None
    track_color: ControlStateValue[ColorValue] = None
    overlay_color: ControlStateValue[ColorValue] = None
    track_outline_color: ControlStateValue[ColorValue] = None
    thumb_icon: ControlStateValue[str] = None
    track_outline_width: ControlStateValue[OptionalNumber] = None
    splash_radius: OptionalNumber = None
    mouse_cursor: ControlStateValue[MouseCursor] = None
    padding: Optional[PaddingValue] = None

    def __post_init__(self):
        if not isinstance(self.thumb_color, dict):
            self.thumb_color = {ControlState.DEFAULT: self.thumb_color}
        if not isinstance(self.track_color, dict):
            self.track_color = {ControlState.DEFAULT: self.track_color}
        if not isinstance(self.overlay_color, dict):
            self.overlay_color = {ControlState.DEFAULT: self.overlay_color}
        if not isinstance(self.track_outline_color, dict):
            self.track_outline_color = {ControlState.DEFAULT: self.track_outline_color}
        if not isinstance(self.thumb_icon, dict):
            self.thumb_icon = {ControlState.DEFAULT: self.thumb_icon}
        if not isinstance(self.track_outline_width, dict):
            self.track_outline_width = {ControlState.DEFAULT: self.track_outline_width}
        if not isinstance(self.mouse_cursor, dict):
            self.mouse_cursor = {ControlState.DEFAULT: self.mouse_cursor}


@dataclass
class DividerTheme:
    color: Optional[ColorValue] = None
    thickness: OptionalNumber = None
    space: OptionalNumber = None
    leading_indent: OptionalNumber = None
    trailing_indent: OptionalNumber = None


@dataclass
class SnackBarTheme:
    bgcolor: Optional[ColorValue] = None
    action_text_color: Optional[ColorValue] = None
    action_bgcolor: Optional[ColorValue] = None
    close_icon_color: Optional[ColorValue] = None
    disabled_action_text_color: Optional[ColorValue] = None
    disabled_action_bgcolor: Optional[ColorValue] = None
    elevation: OptionalNumber = None
    content_text_style: Optional[TextStyle] = None
    width: OptionalNumber = None
    alignment: Optional[Alignment] = None
    show_close_icon: Optional[bool] = None
    dismiss_direction: Optional[DismissDirection] = None
    behavior: Optional[SnackBarBehavior] = None
    shape: Optional[OutlinedBorder] = None
    inset_padding: Optional[PaddingValue] = None
    action_overflow_threshold: OptionalNumber = None


@dataclass
class BannerTheme:
    bgcolor: Optional[ColorValue] = None
    surface_tint_color: Optional[ColorValue] = None
    shadow_color: Optional[ColorValue] = None
    divider_color: Optional[ColorValue] = None
    padding: Optional[PaddingValue] = None
    leading_padding: Optional[PaddingValue] = None
    elevation: OptionalNumber = None
    content_text_style: Optional[TextStyle] = None


@dataclass
class DatePickerTheme:
    bgcolor: Optional[ColorValue] = None
    surface_tint_color: Optional[ColorValue] = None
    shadow_color: Optional[ColorValue] = None
    divider_color: Optional[ColorValue] = None
    header_bgcolor: Optional[ColorValue] = None
    today_bgcolor: ControlStateValue[ColorValue] = None
    day_bgcolor: ControlStateValue[ColorValue] = None
    day_overlay_color: ControlStateValue[ColorValue] = None
    day_foreground_color: ControlStateValue[ColorValue] = None
    elevation: OptionalNumber = None
    range_picker_elevation: OptionalNumber = None
    day_text_style: Optional[TextStyle] = None
    weekday_text_style: Optional[TextStyle] = None
    year_text_style: Optional[TextStyle] = None
    shape: Optional[OutlinedBorder] = None
    cancel_button_style: Optional[ButtonStyle] = None
    confirm_button_style: Optional[ButtonStyle] = None
    header_foreground_color: Optional[ColorValue] = None
    header_headline_text_style: Optional[TextStyle] = None
    header_help_text_style: Optional[TextStyle] = None
    range_picker_bgcolor: Optional[ColorValue] = None
    range_picker_header_bgcolor: Optional[ColorValue] = None
    range_picker_header_foreground_color: Optional[ColorValue] = None
    today_foreground_color: ControlStateValue[ColorValue] = None
    range_picker_shape: Optional[OutlinedBorder] = None
    range_picker_header_help_text_style: Optional[TextStyle] = None
    range_picker_header_headline_text_style: Optional[TextStyle] = None
    range_picker_surface_tint_color: Optional[ColorValue] = None
    range_selection_bgcolor: Optional[ColorValue] = None
    range_selection_overlay_color: ControlStateValue[ColorValue] = None
    today_border_side: Optional[BorderSide] = None
    year_bgcolor: ControlStateValue[ColorValue] = None
    year_foreground_color: ControlStateValue[ColorValue] = None
    year_overlay_color: ControlStateValue[ColorValue] = None
    day_shape: ControlStateValue[OutlinedBorder] = None
    locale: Optional[Locale] = None

    def __post_init__(self):
        if not isinstance(self.today_bgcolor, dict):
            self.today_bgcolor = {ControlState.DEFAULT: self.today_bgcolor}
        if not isinstance(self.day_bgcolor, dict):
            self.day_bgcolor = {ControlState.DEFAULT: self.day_bgcolor}
        if not isinstance(self.day_overlay_color, dict):
            self.day_overlay_color = {ControlState.DEFAULT: self.day_overlay_color}
        if not isinstance(self.day_foreground_color, dict):
            self.day_foreground_color = {
                ControlState.DEFAULT: self.day_foreground_color
            }
        if not isinstance(self.today_foreground_color, dict):
            self.today_foreground_color = {
                ControlState.DEFAULT: self.today_foreground_color
            }
        if not isinstance(self.range_selection_overlay_color, dict):
            self.range_selection_overlay_color = {
                ControlState.DEFAULT: self.range_selection_overlay_color
            }
        if not isinstance(self.year_bgcolor, dict):
            self.year_bgcolor = {ControlState.DEFAULT: self.year_bgcolor}
        if not isinstance(self.year_foreground_color, dict):
            self.year_foreground_color = {
                ControlState.DEFAULT: self.year_foreground_color
            }
        if not isinstance(self.year_overlay_color, dict):
            self.year_overlay_color = {ControlState.DEFAULT: self.year_overlay_color}
        if not isinstance(self.day_shape, dict):
            self.day_shape = {ControlState.DEFAULT: self.day_shape}


@dataclass
class TimePickerTheme:
    bgcolor: Optional[ColorValue] = None
    day_period_color: Optional[ColorValue] = None
    day_period_text_color: Optional[ColorValue] = None
    dial_bgcolor: Optional[ColorValue] = None
    dial_hand_color: Optional[ColorValue] = None
    dial_text_color: Optional[ColorValue] = None
    entry_mode_icon_color: Optional[ColorValue] = None
    hour_minute_color: Optional[ColorValue] = None
    hour_minute_text_color: Optional[ColorValue] = None
    day_period_button_style: Optional[ButtonStyle] = None
    cancel_button_style: Optional[ButtonStyle] = None
    confirm_button_style: Optional[ButtonStyle] = None
    day_period_text_style: Optional[TextStyle] = None
    dial_text_style: Optional[TextStyle] = None
    help_text_style: Optional[TextStyle] = None
    hour_minute_text_style: Optional[TextStyle] = None
    elevation: OptionalNumber = None
    shape: Optional[OutlinedBorder] = None
    day_period_shape: Optional[OutlinedBorder] = None
    hour_minute_shape: Optional[OutlinedBorder] = None
    day_period_border_side: Optional[BorderSide] = None
    padding: Optional[PaddingValue] = None
    time_selector_separator_color: ControlStateValue[ColorValue] = None
    time_selector_separator_text_style: ControlStateValue[TextStyle] = None

    def __post_init__(self):
        if not isinstance(self.time_selector_separator_color, dict):
            self.time_selector_separator_color = {
                ControlState.DEFAULT: self.time_selector_separator_color
            }
        if not isinstance(self.time_selector_separator_text_style, dict):
            self.time_selector_separator_text_style = {
                ControlState.DEFAULT: self.time_selector_separator_text_style
            }


@dataclass
class DropdownMenuTheme:
    menu_style: Optional[MenuStyle] = None
    text_style: Optional[TextStyle] = None


@dataclass
class ListTileTheme:
    icon_color: Optional[ColorValue] = None
    text_color: Optional[ColorValue] = None
    bgcolor: Optional[ColorValue] = None
    selected_tile_color: Optional[ColorValue] = None
    selected_color: Optional[ColorValue] = None
    is_three_line: Optional[bool] = None
    enable_feedback: Optional[bool] = None
    dense: Optional[bool] = None
    shape: Optional[OutlinedBorder] = None
    visual_density: Optional[VisualDensity] = None
    content_padding: Optional[PaddingValue] = None
    min_vertical_padding: Optional[PaddingValue] = None
    horizontal_spacing: OptionalNumber = None
    min_leading_width: OptionalNumber = None
    title_text_style: Optional[TextStyle] = None
    subtitle_text_style: Optional[TextStyle] = None
    leading_and_trailing_text_style: Optional[TextStyle] = None
    mouse_cursor: ControlStateValue[MouseCursor] = None
    min_tile_height: OptionalNumber = None

    def __post_init__(self):
        if not isinstance(self.mouse_cursor, dict):
            self.mouse_cursor = {ControlState.DEFAULT: self.mouse_cursor}


@dataclass
class TooltipTheme:
    height: OptionalNumber = None
    text_style: Optional[TextStyle] = None
    enable_feedback: Optional[bool] = None
    exclude_from_semantics: Optional[bool] = None
    prefer_below: Optional[bool] = None
    vertical_offset: OptionalNumber = None
    padding: Optional[PaddingValue] = None
    wait_duration: Optional[DurationValue] = None
    exit_duration: Optional[DurationValue] = None
    show_duration: Optional[DurationValue] = None
    margin: Optional[MarginValue] = None
    trigger_mode: Optional[TooltipTriggerMode] = None
    decoration: Optional[BoxDecoration] = None
    text_align: Optional[TextAlign] = None


@dataclass
class ExpansionTileTheme:
    bgcolor: Optional[ColorValue] = None
    icon_color: Optional[ColorValue] = None
    text_color: Optional[ColorValue] = None
    collapsed_bgcolor: Optional[ColorValue] = None
    collapsed_icon_color: Optional[ColorValue] = None
    clip_behavior: Optional[ClipBehavior] = None
    collapsed_text_color: Optional[ColorValue] = None
    tile_padding: Optional[PaddingValue] = None
    expanded_alignment: Optional[Alignment] = None
    controls_padding: Optional[PaddingValue] = None


@dataclass
class SliderTheme:
    active_track_color: Optional[ColorValue] = None
    inactive_track_color: Optional[ColorValue] = None
    thumb_color: Optional[ColorValue] = None
    overlay_color: Optional[ColorValue] = None
    value_indicator_color: Optional[ColorValue] = None
    disabled_thumb_color: Optional[ColorValue] = None
    value_indicator_text_style: Optional[TextStyle] = None
    mouse_cursor: ControlStateValue[MouseCursor] = None
    active_tick_mark_color: Optional[ColorValue] = None
    disabled_active_tick_mark_color: Optional[ColorValue] = None
    disabled_active_track_color: Optional[ColorValue] = None
    disabled_inactive_tick_mark_color: Optional[ColorValue] = None
    disabled_inactive_track_color: Optional[ColorValue] = None
    disabled_secondary_active_track_color: Optional[ColorValue] = None
    inactive_tick_mark_color: Optional[ColorValue] = None
    overlapping_shape_stroke_color: Optional[ColorValue] = None
    min_thumb_separation: OptionalNumber = None
    secondary_active_track_color: Optional[ColorValue] = None
    track_height: OptionalNumber = None
    value_indicator_stroke_color: Optional[ColorValue] = None
    interaction: Optional[SliderInteraction] = None
    padding: Optional[PaddingValue] = None
    track_gap: OptionalNumber = None
    thumb_size: ControlStateValue[Size] = None
    year_2023: Optional[bool] = None

    def __post_init__(self):
        if not isinstance(self.mouse_cursor, dict):
            self.mouse_cursor = {ControlState.DEFAULT: self.mouse_cursor}
        if not isinstance(self.thumb_size, dict):
            self.thumb_size = {ControlState.DEFAULT: self.thumb_size}
        if self.year_2023 is not None:
            deprecated_property(
                name="year_2023",
                version="0.27.0",
                delete_version=None,  # not known for now
                reason="Set this flag to False to opt into the 2024 Slider appearance. In the future, this flag will default to False.",
            )


@dataclass
class ProgressIndicatorTheme:
    color: Optional[ColorValue] = None
    circular_track_color: Optional[ColorValue] = None
    linear_track_color: Optional[ColorValue] = None
    refresh_bgcolor: Optional[ColorValue] = None
    linear_min_height: OptionalNumber = None
    border_radius: Optional[BorderRadius] = None
    track_gap: OptionalNumber = None
    circular_track_padding: Optional[PaddingValue] = None
    size_constraints: Optional[BoxConstraints] = None
    stop_indicator_color: Optional[ColorValue] = None
    stop_indicator_radius: OptionalNumber = None
    stroke_align: OptionalNumber = None
    stroke_cap: Optional[StrokeCap] = None
    stroke_width: OptionalNumber = None
    year_2023: Optional[bool] = None

    def __post_init__(self):
        if self.year_2023 is not None:
            deprecated_property(
                name="year_2023",
                version="0.27.0",
                delete_version=None,  # not known for now
                reason="Set this flag to False to opt into the 2024 ProgressIndicator appearance. In the future, this flag will default to False.",
            )


@dataclass
class PopupMenuTheme:
    color: Optional[ColorValue] = None
    surface_tint_color: Optional[ColorValue] = None
    shadow_color: Optional[ColorValue] = None
    icon_color: Optional[ColorValue] = None
    text_style: Optional[TextStyle] = None
    label_text_style: Optional[TextStyle] = None
    enable_feedback: Optional[bool] = None
    elevation: OptionalNumber = None
    icon_size: OptionalNumber = None
    shape: Optional[OutlinedBorder] = None
    menu_position: Optional[PopupMenuPosition] = None
    mouse_cursor: ControlStateValue[MouseCursor] = None
    menu_padding: Optional[PaddingValue] = None

    def __post_init__(self):
        if not isinstance(self.mouse_cursor, dict):
            self.mouse_cursor = {ControlState.DEFAULT: self.mouse_cursor}
        if not isinstance(self.thumb_size, dict):
            self.thumb_size = {ControlState.DEFAULT: self.thumb_size}
        if self.year_2023 is not None:
            deprecated_property(
                name="year_2023",
                version="0.27.0",
                delete_version=None,  # not known for now
                reason="Set this flag to False to opt into the 2024 Slider appearance. In the future, this flag will default to False.",
            )


@dataclass
class SearchBarTheme:
    bgcolor: Optional[ColorValue] = None
    text_capitalization: Optional[TextCapitalization] = None
    shadow_color: ControlStateValue[ColorValue] = None
    surface_tint_color: ControlStateValue[ColorValue] = None
    overlay_color: ControlStateValue[ColorValue] = None
    elevation: ControlStateValue[OptionalNumber] = None
    text_style: ControlStateValue[TextStyle] = None
    hint_style: ControlStateValue[TextStyle] = None
    shape: ControlStateValue[OutlinedBorder] = None
    padding: ControlStateValue[PaddingValue] = None
    size_constraints: Optional[BoxConstraints] = None
    border_side: ControlStateValue[BorderSide] = None

    def __post_init__(self):
        if not isinstance(self.shadow_color, dict):
            self.shadow_color = {ControlState.DEFAULT: self.shadow_color}
        if not isinstance(self.surface_tint_color, dict):
            self.surface_tint_color = {ControlState.DEFAULT: self.surface_tint_color}
        if not isinstance(self.overlay_color, dict):
            self.overlay_color = {ControlState.DEFAULT: self.overlay_color}
        if not isinstance(self.elevation, dict):
            self.elevation = {ControlState.DEFAULT: self.elevation}
        if not isinstance(self.text_style, dict):
            self.text_style = {ControlState.DEFAULT: self.text_style}
        if not isinstance(self.hint_style, dict):
            self.hint_style = {ControlState.DEFAULT: self.hint_style}
        if not isinstance(self.shape, dict):
            self.shape = {ControlState.DEFAULT: self.shape}
        if not isinstance(self.padding, dict):
            self.padding = {ControlState.DEFAULT: self.padding}
        if not isinstance(self.border_side, dict):
            self.border_side = {ControlState.DEFAULT: self.border_side}


@dataclass
class SearchViewTheme:
    bgcolor: Optional[ColorValue] = None
    surface_tint_color: Optional[ColorValue] = None
    divider_color: Optional[ColorValue] = None
    elevation: OptionalNumber = None
    header_hint_text_style: Optional[TextStyle] = None
    header_text_style: Optional[TextStyle] = None
    shape: Optional[OutlinedBorder] = None
    border_side: Optional[BorderSide] = None
    size_constraints: Optional[BoxConstraints] = None
    header_height: OptionalNumber = None
    padding: Optional[PaddingValue] = None
    bar_padding: Optional[PaddingValue] = None
    shrink_wrap: Optional[bool] = None


@dataclass
class NavigationDrawerTheme:
    bgcolor: Optional[ColorValue] = None
    shadow_color: Optional[ColorValue] = None
    surface_tint_color: Optional[ColorValue] = None
    indicator_color: Optional[ColorValue] = None
    elevation: OptionalNumber = None
    tile_height: OptionalNumber = None
    label_text_style: ControlStateValue[TextStyle] = None
    indicator_shape: Optional[OutlinedBorder] = None
    indicator_size: Optional[Size] = None

    def __post_init__(self):
        if not isinstance(self.label_text_style, dict):
            self.label_text_style = {ControlState.DEFAULT: self.label_text_style}


@dataclass
class NavigationBarTheme:
    bgcolor: Optional[ColorValue] = None
    shadow_color: Optional[ColorValue] = None
    surface_tint_color: Optional[ColorValue] = None
    indicator_color: Optional[ColorValue] = None
    overlay_color: ControlStateValue[ColorValue] = None
    elevation: OptionalNumber = None
    height: OptionalNumber = None
    label_text_style: ControlStateValue[TextStyle] = None
    indicator_shape: Optional[OutlinedBorder] = None
    label_behavior: Optional[NavigationBarLabelBehavior] = None
    label_padding: Optional[PaddingValue] = None

    def __post_init__(self):
        if not isinstance(self.label_text_style, dict):
            self.label_text_style = {ControlState.DEFAULT: self.label_text_style}
        if not isinstance(self.overlay_color, dict):
            self.overlay_color = {ControlState.DEFAULT: self.overlay_color}


@dataclass
class SegmentedButtonTheme:
    selected_icon: Optional[IconValue] = None
    style: Optional[ButtonStyle] = None


@dataclass
class IconTheme:
    color: Optional[ColorValue] = None
    apply_text_scaling: Optional[bool] = None
    fill: OptionalNumber = None
    opacity: OptionalNumber = None
    size: OptionalNumber = None
    optical_size: OptionalNumber = None
    grade: OptionalNumber = None
    weight: OptionalNumber = None
    shadows: Optional[List[BoxShadow]] = None


@dataclass
class DataTableTheme:
    checkbox_horizontal_margin: OptionalNumber = None
    column_spacing: OptionalNumber = None
    data_row_max_height: OptionalNumber = None
    data_row_min_height: OptionalNumber = None
    data_row_color: ControlStateValue[ColorValue] = None
    data_text_style: Optional[TextStyle] = None
    divider_thickness: OptionalNumber = None
    horizontal_margin: OptionalNumber = None
    heading_text_style: Optional[TextStyle] = None
    heading_row_color: ControlStateValue[ColorValue] = None
    heading_row_height: OptionalNumber = None
    data_row_cursor: ControlStateValue[MouseCursor] = None
    decoration: Optional[BoxDecoration] = None
    heading_row_alignment: Optional[MainAxisAlignment] = None
    heading_cell_cursor: ControlStateValue[MouseCursor] = None


@deprecated_class(
    "Use ElevatedButtonTheme, OutlinedButtonTheme, TextButtonTheme, FilledButtonTheme or IconButtonTheme instead.",
    version="0.27.0",
    delete_version="0.30.0",
)
@dataclass
class ButtonTheme:
    button_color: Optional[ColorValue] = None
    disabled_color: Optional[ColorValue] = None
    hover_color: Optional[ColorValue] = None
    focus_color: Optional[ColorValue] = None
    highlight_color: Optional[ColorValue] = None
    splash_color: Optional[ColorValue] = None
    color_scheme: Optional[ColorScheme] = None
    aligned_dropdown: Optional[bool] = None
    height: OptionalNumber = None
    min_width: OptionalNumber = None
    shape: Optional[OutlinedBorder] = None
    padding: Optional[PaddingValue] = None


@dataclass
class Theme:
    color_scheme_seed: Optional[ColorValue] = None
    primary_swatch: Optional[ColorValue] = None
    font_family: Optional[str] = None
    use_material3: Optional[bool] = None
    appbar_theme: Optional[AppBarTheme] = None
    badge_theme: Optional[BadgeTheme] = None
    banner_theme: Optional[BannerTheme] = None
    bottom_appbar_theme: Optional[BottomAppBarTheme] = None
    bottom_sheet_theme: Optional[BottomSheetTheme] = None
    button_theme: Optional[ButtonTheme] = None
    card_theme: Optional[CardTheme] = None
    checkbox_theme: Optional[CheckboxTheme] = None
    chip_theme: Optional[ChipTheme] = None
    color_scheme: Optional[ColorScheme] = None
    data_table_theme: Optional[DataTableTheme] = None
    date_picker_theme: Optional[DatePickerTheme] = None
    dialog_theme: Optional[DialogTheme] = None
    divider_theme: Optional[DividerTheme] = None
    # dropdown_menu_theme: Optional[DropdownMenuTheme] = None
    elevated_button_theme: Optional[ElevatedButtonTheme] = None
    outlined_button_theme: Optional[OutlinedButtonTheme] = None
    text_button_theme: Optional[TextButtonTheme] = None
    filled_button_theme: Optional[FilledButtonTheme] = None
    icon_button_theme: Optional[IconButtonTheme] = None
    expansion_tile_theme: Optional[ExpansionTileTheme] = None
    floating_action_button_theme: Optional[FloatingActionButtonTheme] = None
    icon_theme: Optional[IconTheme] = None
    list_tile_theme: Optional[ListTileTheme] = None
    navigation_bar_theme: Optional[NavigationBarTheme] = None
    navigation_drawer_theme: Optional[NavigationDrawerTheme] = None
    navigation_rail_theme: Optional[NavigationRailTheme] = None
    page_transitions: PageTransitionsTheme = field(default_factory=PageTransitionsTheme)
    popup_menu_theme: Optional[PopupMenuTheme] = None
    splash_color: Optional[ColorValue] = None
    highlight_color: Optional[ColorValue] = None
    hover_color: Optional[ColorValue] = None
    focus_color: Optional[ColorValue] = None
    unselected_control_color: Optional[ColorValue] = None
    disabled_color: Optional[ColorValue] = None
    canvas_color: Optional[ColorValue] = None
    scaffold_bgcolor: Optional[ColorValue] = None
    card_color: Optional[ColorValue] = None
    divider_color: Optional[ColorValue] = None
    dialog_bgcolor: Optional[ColorValue] = None
    indicator_color: Optional[ColorValue] = None
    hint_color: Optional[ColorValue] = None
    shadow_color: Optional[ColorValue] = None
    secondary_header_color: Optional[ColorValue] = None
    primary_color: Optional[ColorValue] = None
    primary_color_dark: Optional[ColorValue] = None
    primary_color_light: Optional[ColorValue] = None
    primary_text_theme: Optional[TextTheme] = None
    progress_indicator_theme: Optional[ProgressIndicatorTheme] = None
    radio_theme: Optional[RadioTheme] = None
    scrollbar_theme: Optional[ScrollbarTheme] = None
    search_bar_theme: Optional[SearchBarTheme] = None
    search_view_theme: Optional[SearchViewTheme] = None
    segmented_button_theme: Optional[SegmentedButtonTheme] = None
    slider_theme: Optional[SliderTheme] = None
    snackbar_theme: Optional[SnackBarTheme] = None
    switch_theme: Optional[SwitchTheme] = None
    system_overlay_style: SystemOverlayStyle = field(default_factory=SystemOverlayStyle)
    tabs_theme: Optional[TabsTheme] = None
    text_theme: Optional[TextTheme] = None
    time_picker_theme: Optional[TimePickerTheme] = None
    tooltip_theme: Optional[TooltipTheme] = None
    visual_density: Optional[VisualDensity] = None

    def __post_init__(self):
        if self.button_theme:
            deprecated_property(
                "button_theme",
                "Use elevated_button_theme, outlined_button_theme, text_button_theme, filled_button_theme or icon_button_theme instead.",
                version="0.27.0",
                delete_version="0.30.0",
            )
        if self.dialog_bgcolor:
            deprecated_property(
                "dialog_bgcolor",
                "Use dialog_theme.bgcolor instead.",
                version="0.27.0",
                delete_version="0.30.0",
            )



import json
import time
from typing import Optional

from flet.core.animation import AnimationCurve
from flet.core.control import Control, OptionalNumber
from flet.core.control_event import ControlEvent
from flet.core.event_handler import EventHandler
from flet.core.types import OptionalEventCallable, ScrollMode


class ScrollableControl(Control):
    def __init__(
        self,
        scroll: Optional[ScrollMode] = None,
        auto_scroll: Optional[bool] = None,
        reverse: Optional[bool] = None,
        on_scroll_interval: OptionalNumber = None,
        on_scroll: OptionalEventCallable["OnScrollEvent"] = None,
    ):
        super().__init__()
        self.__on_scroll = EventHandler(lambda e: OnScrollEvent(e))
        self._add_event_handler("onScroll", self.__on_scroll.get_handler())

        self.scroll = scroll
        self.auto_scroll = auto_scroll
        self.reverse = reverse
        self.on_scroll_interval = on_scroll_interval
        self.on_scroll = on_scroll

    def scroll_to(
        self,
        offset: Optional[float] = None,
        delta: Optional[float] = None,
        key: Optional[str] = None,
        duration: Optional[int] = None,
        curve: Optional[AnimationCurve] = None,
    ):
        m = {
            "n": "scroll_to",
            "i": str(time.time()),
            "p": {
                "offset": offset,
                "delta": delta,
                "key": key,
                "duration": duration,
                "curve": curve.value if curve is not None else None,
            },
        }
        self._set_attr_json("method", m)
        self.update()

    # scroll
    @property
    def scroll(self) -> Optional[ScrollMode]:
        return self.__scroll

    @scroll.setter
    def scroll(self, value: Optional[ScrollMode]):
        self.__scroll = value
        self._set_attr(
            "scroll",
            value.value
            if isinstance(value, ScrollMode)
            else "auto"
            if value is True
            else None
            if value is False
            else value,
        )

    # auto_scroll
    @property
    def auto_scroll(self) -> bool:
        return self._get_attr("autoScroll", data_type="bool", def_value=False)

    @auto_scroll.setter
    def auto_scroll(self, value: Optional[bool]):
        self._set_attr("autoScroll", value)

    # reverse
    @property
    def reverse(self) -> bool:
        return self._get_attr("reverse", data_type="bool", def_value=False)

    @reverse.setter
    def reverse(self, value: Optional[bool]):
        self._set_attr("reverse", value)

    # on_scroll_interval
    @property
    def on_scroll_interval(self) -> OptionalNumber:
        return self._get_attr("onScrollInterval")

    @on_scroll_interval.setter
    def on_scroll_interval(self, value: OptionalNumber):
        self._set_attr("onScrollInterval", value)

    # on_scroll
    @property
    def on_scroll(self) -> OptionalEventCallable["OnScrollEvent"]:
        return self.__on_scroll.handler

    @on_scroll.setter
    def on_scroll(self, handler: OptionalEventCallable["OnScrollEvent"]):
        self.__on_scroll.handler = handler
        self._set_attr("onScroll", True if handler is not None else None)


class OnScrollEvent(ControlEvent):
    def __init__(self, e: ControlEvent):
        super().__init__(e.target, e.name, e.data, e.control, e.page)
        d = json.loads(e.data)
        self.event_type: str = d.get("t")
        self.pixels: float = d.get("p")
        self.min_scroll_extent: float = d.get("minse")
        self.max_scroll_extent: float = d.get("maxse")
        self.viewport_dimension: float = d.get("vd")
        self.scroll_delta: Optional[float] = d.get("sd")
        self.direction: Optional[str] = d.get("dir")
        self.overscroll: Optional[float] = d.get("os")
        self.velocity: Optional[float] = d.get("v")



import asyncio
import json
import logging
import os
import struct
import sys
import tempfile
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import List, Optional

import flet
from flet.core.local_connection import LocalConnection
from flet.core.protocol import (
    ClientActions,
    ClientMessage,
    Command,
    CommandEncoder,
    PageCommandResponsePayload,
    PageCommandsBatchResponsePayload,
    RegisterWebClientRequestPayload,
)
from flet.core.pubsub.pubsub_hub import PubSubHub
from flet.utils import get_free_tcp_port, is_windows, random_string

logger = logging.getLogger(flet.__name__)


class FletSocketServer(LocalConnection):
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        port: int = 0,
        uds_path: Optional[str] = None,
        on_event=None,
        on_session_created=None,
        blocking=False,
        executor: Optional[ThreadPoolExecutor] = None,
    ):
        super().__init__()
        self.__send_queue = asyncio.Queue()
        self.__port = port
        self.__uds_path = uds_path
        self.__on_event = on_event
        self.__on_session_created = on_session_created
        self.__blocking = blocking
        self.__loop = loop
        self.__executor = executor
        self.pubsubhub = PubSubHub(loop=loop, executor=executor)
        self.__running_tasks = set()

    async def start(self):
        self.__connected = False
        self.__receive_loop_task = None
        self.__send_loop_task = None
        if is_windows() or self.__port > 0:
            # TCP
            host = "localhost"
            port = self.__port if self.__port > 0 else get_free_tcp_port()
            self.page_url = f"tcp://{host}:{port}"
            logger.info(f"Starting up TCP server on {host}:{port}")
            server = await asyncio.start_server(self.handle_connection, host, port)
        else:
            # UDS
            if not self.__uds_path:
                self.__uds_path = str(
                    Path(tempfile.gettempdir()).joinpath(random_string(10))
                )
            if os.path.exists(self.__uds_path):
                os.remove(self.__uds_path)
            self.page_url = self.__uds_path
            logger.info(f"Starting up UDS server on {self.__uds_path}")
            server = await asyncio.start_unix_server(
                self.handle_connection, self.__uds_path
            )

        if self.__blocking:
            self.__server = None
            await server.serve_forever()
        else:
            self.__server = asyncio.create_task(server.serve_forever())

    async def handle_connection(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        if not self.__connected:
            self.__connected = True
            logger.debug("Connected new TCP client")
            self.__receive_loop_task = asyncio.create_task(self.__receive_loop(reader))
            self.__send_loop_task = asyncio.create_task(self.__send_loop(writer))

    async def __receive_loop(self, reader: asyncio.StreamReader):
        while True:
            try:
                raw_msglen = await reader.readexactly(4)
            except Exception:
                return None

            if not raw_msglen:
                return None
            msglen = struct.unpack(">I", raw_msglen)[0]

            data = await reader.readexactly(msglen)
            await self.__on_message(data.decode("utf-8"))

    async def __send_loop(self, writer: asyncio.StreamWriter):
        while True:
            message = await self.__send_queue.get()
            try:
                data = message.encode("utf-8")
                msg = struct.pack(">I", len(data)) + data
                writer.write(msg)
                # await writer.drain()
                logger.debug(f"sent to TCP: {len(msg)}")
            except Exception:
                # re-enqueue the message to repeat it when re-connected
                self.__send_queue.put_nowait(message)
                raise

    async def __on_message(self, data: str):
        logger.debug(f"_on_message: {data}")
        msg_dict = json.loads(data)
        msg = ClientMessage(**msg_dict)
        task = None
        if msg.action == ClientActions.REGISTER_WEB_CLIENT:
            self._client_details = RegisterWebClientRequestPayload(**msg.payload)

            # register response
            self.__send(self._create_register_web_client_response())

            # start session
            if self.__on_session_created is not None:
                task = asyncio.create_task(
                    self.__on_session_created(self._create_session_handler_arg())
                )

        elif msg.action == ClientActions.PAGE_EVENT_FROM_WEB:
            if self.__on_event is not None:
                task = asyncio.create_task(
                    self.__on_event(self._create_page_event_handler_arg(msg))
                )

        elif msg.action == ClientActions.UPDATE_CONTROL_PROPS:
            if self.__on_event is not None:
                task = asyncio.create_task(
                    self.__on_event(self._create_update_control_props_handler_arg(msg))
                )
        else:
            # it's something else
            raise Exception(f'Unknown message "{msg.action}": {msg.payload}')

        if task:
            self.__running_tasks.add(task)
            task.add_done_callback(self.__running_tasks.discard)

    def send_command(self, session_id: str, command: Command):
        result, message = self._process_command(command)
        if message:
            self.__send(message)
        return PageCommandResponsePayload(result=result, error="")

    def send_commands(self, session_id: str, commands: List[Command]):
        results = []
        messages = []
        for command in commands:
            result, message = self._process_command(command)
            if command.name in ["add", "get"]:
                results.append(result)
            if message:
                messages.append(message)
        if len(messages) > 0:
            self.__send(ClientMessage(ClientActions.PAGE_CONTROLS_BATCH, messages))
        return PageCommandsBatchResponsePayload(results=results, error="")

    def __send(self, message: ClientMessage):
        j = json.dumps(message, cls=CommandEncoder, separators=(",", ":"))
        logger.debug(f"__send: {j}")
        self.__loop.call_soon_threadsafe(self.__send_queue.put_nowait, j)

    async def close(self):
        logger.debug("Closing connection...")

        logger.debug(f"Disconnecting all pages...")
        while self.sessions:
            _, page = self.sessions.popitem()
            await page._disconnect(0)

        if self.__executor:
            logger.debug("Shutting down thread pool...")
            if sys.version_info >= (3, 9):
                self.__executor.shutdown(wait=False, cancel_futures=True)
            else:
                self.__executor.shutdown(wait=False)

        # close socket
        if self.__receive_loop_task:
            self.__receive_loop_task.cancel()
        if self.__send_loop_task:
            self.__send_loop_task.cancel()
        if self.__server:
            self.__server.cancel()

        # remove UDS path
        if self.__uds_path and os.path.exists(self.__uds_path):
            os.unlink(self.__uds_path)



from datetime import datetime, time
from enum import Enum
from typing import Any, Optional, Union

from flet.core.control import Control, OptionalNumber
from flet.core.control_event import ControlEvent
from flet.core.event_handler import EventHandler
from flet.core.ref import Ref
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    ColorEnums,
    ColorValue,
    OptionalControlEventCallable,
    OptionalEventCallable,
    Orientation,
    ResponsiveNumber,
)


class TimePickerEntryMode(Enum):
    DIAL = "dial"
    INPUT = "input"
    DIAL_ONLY = "dialOnly"
    INPUT_ONLY = "inputOnly"


class TimePickerEntryModeChangeEvent(ControlEvent):
    def __init__(self, e: ControlEvent):
        super().__init__(e.target, e.name, e.data, e.control, e.page)
        self.entry_mode: Optional[TimePickerEntryMode] = TimePickerEntryMode(e.data)


class TimePicker(Control):
    """
    A Material-style time picker dialog.

    It is added to [`page.overlay`](page#overlay) and can be opened by setting `open=True` or by calling `Page.open()` method.

    Depending on the `time_picker_entry_mode`, it will show either a Dial or an Input (hour and minute text fields) for picking a time.

    Example:
    ```
    import flet as ft


    def main(page: ft.Page):
        page.horizontal_alignment = ft.CrossAxisAlignment.CENTER

        def handle_change(e):
            page.add(ft.Text(f"TimePicker change: {time_picker.value}"))

        def handle_dismissal(e):
            page.add(ft.Text(f"TimePicker dismissed: {time_picker.value}"))

        def handle_entry_mode_change(e):
            page.add(ft.Text(f"TimePicker Entry mode changed to {e.entry_mode}"))

        time_picker = ft.TimePicker(
            confirm_text="Confirm",
            error_invalid_text="Time out of range",
            help_text="Pick your time slot",
            on_change=handle_change,
            on_dismiss=handle_dismissal,
            on_entry_mode_change=handle_entry_mode_change,
        )

        page.add(
            ft.ElevatedButton(
                "Pick time",
                icon=ft.icons.TIME_TO_LEAVE,
                on_click=lambda _: page.open(time_picker),
            )
        )


    ft.app(main)
    ```

    -----

    Online docs: https://flet.dev/docs/controls/time_picker
    """

    def __init__(
        self,
        value: Optional[time] = datetime.now().time(),
        open: bool = False,
        time_picker_entry_mode: Optional[TimePickerEntryMode] = None,
        hour_label_text: Optional[str] = None,
        minute_label_text: Optional[str] = None,
        help_text: Optional[str] = None,
        cancel_text: Optional[str] = None,
        confirm_text: Optional[str] = None,
        error_invalid_text: Optional[str] = None,
        orientation: Optional[Orientation] = None,
        barrier_color: Optional[ColorValue] = None,
        on_change: OptionalControlEventCallable = None,
        on_dismiss: OptionalControlEventCallable = None,
        on_entry_mode_change: OptionalEventCallable[
            TimePickerEntryModeChangeEvent
        ] = None,
        #
        # Control
        #
        ref: Optional[Ref] = None,
        expand: Optional[Union[bool, int]] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        tooltip: Optional[TooltipValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        Control.__init__(
            self,
            ref=ref,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            tooltip=tooltip,
            visible=visible,
            disabled=disabled,
            data=data,
        )

        self.__on_entry_mode_change = EventHandler(
            lambda e: TimePickerEntryModeChangeEvent(e)
        )
        self._add_event_handler(
            "entryModeChange", self.__on_entry_mode_change.get_handler()
        )

        self.value = value
        self.help_text = help_text
        self.cancel_text = cancel_text
        self.confirm_text = confirm_text
        self.error_invalid_text = error_invalid_text
        self.hour_label_text = hour_label_text
        self.minute_label_text = minute_label_text
        self.time_picker_entry_mode = time_picker_entry_mode
        self.orientation = orientation
        self.on_change = on_change
        self.on_dismiss = on_dismiss
        self.open = open
        self.on_entry_mode_change = on_entry_mode_change
        self.barrier_color = barrier_color

    def _get_control_name(self):
        return "timepicker"

    # open
    @property
    def open(self) -> bool:
        return self._get_attr("open", data_type="bool", def_value=False)

    @open.setter
    def open(self, value: Optional[bool]):
        self._set_attr("open", value)

    # value
    @property
    def value(self) -> time:
        v = self._get_attr("value")  # format HH:MM
        return time(*map(int, v.split(":")))

    @value.setter
    def value(self, value: time):
        self._set_attr("value", value.strftime("%H:%M"))

    # hour_label_text
    @property
    def hour_label_text(self) -> Optional[str]:
        return self._get_attr("hourLabelText", def_value=None)

    @hour_label_text.setter
    def hour_label_text(self, value: Optional[str]):
        self._set_attr("hourLabelText", value)

    # minute_label_text
    @property
    def minute_label_text(self) -> Optional[str]:
        return self._get_attr("minuteLabelText", def_value=None)

    @minute_label_text.setter
    def minute_label_text(self, value: Optional[str]):
        self._set_attr("minuteLabelText", value)

    # help_text
    @property
    def help_text(self) -> Optional[str]:
        return self._get_attr("helpText", def_value=None)

    @help_text.setter
    def help_text(self, value: Optional[str]):
        self._set_attr("helpText", value)

    # cancel_text
    @property
    def cancel_text(self) -> Optional[str]:
        return self._get_attr("cancelText", def_value=None)

    @cancel_text.setter
    def cancel_text(self, value: Optional[str]):
        self._set_attr("cancelText", value)

    # confirm_text
    @property
    def confirm_text(self) -> Optional[str]:
        return self._get_attr("confirmText", def_value=None)

    @confirm_text.setter
    def confirm_text(self, value: Optional[str]):
        self._set_attr("confirmText", value)

    # error_invalid_text
    @property
    def error_invalid_text(self) -> Optional[str]:
        return self._get_attr("errorInvalidText", def_value=None)

    @error_invalid_text.setter
    def error_invalid_text(self, value: Optional[str]):
        self._set_attr("errorInvalidText", value)

    # time_picker_entry_mode
    @property
    def time_picker_entry_mode(self) -> Optional[TimePickerEntryMode]:
        return self.__time_picker_entry_mode

    @time_picker_entry_mode.setter
    def time_picker_entry_mode(self, value: Optional[TimePickerEntryMode]):
        self.__time_picker_entry_mode = value
        self._set_enum_attr("timePickerEntryMode", value, TimePickerEntryMode)

    # orientation
    @property
    def orientation(self) -> Optional[Orientation]:
        return self.__orientation

    @orientation.setter
    def orientation(self, value: Optional[Orientation]):
        self.__orientation = value
        self._set_enum_attr("orientation", value, Orientation)

    # on_change
    @property
    def on_change(self) -> OptionalControlEventCallable:
        return self._get_event_handler("change")

    @on_change.setter
    def on_change(self, handler: OptionalControlEventCallable):
        self._add_event_handler("change", handler)

    # on_dismiss
    @property
    def on_dismiss(self) -> OptionalControlEventCallable:
        return self._get_event_handler("dismiss")

    @on_dismiss.setter
    def on_dismiss(self, handler: OptionalControlEventCallable):
        self._add_event_handler("dismiss", handler)

    # on_entry_mode_change
    @property
    def on_entry_mode_change(
        self,
    ) -> OptionalEventCallable[TimePickerEntryModeChangeEvent]:
        return self.__on_entry_mode_change.handler

    @on_entry_mode_change.setter
    def on_entry_mode_change(
        self, handler: OptionalEventCallable[TimePickerEntryModeChangeEvent]
    ):
        self.__on_entry_mode_change.handler = handler

    # barrier_color
    @property
    def barrier_color(self) -> Optional[ColorValue]:
        return self.__barrier_color

    @barrier_color.setter
    def barrier_color(self, value: Optional[ColorValue]):
        self.__barrier_color = value
        self._set_enum_attr("barrierColor", value, ColorEnums)



from typing import Any, Optional

from flet.core.control import Control, OptionalNumber
from flet.core.ref import Ref
from flet.core.text_style import TextStyle
from flet.core.types import OptionalControlEventCallable


class CupertinoDialogAction(Control):
    """
    A button typically used in a CupertinoAlertDialog.

    Example:
    ```
    import flet as ft


    def main(page: ft.Page):
        page.horizontal_alignment = ft.CrossAxisAlignment.CENTER

        def dialog_dismissed(e):
            page.add(ft.Text("Dialog dismissed"))

        def handle_action_click(e):
            page.add(ft.Text(f"Action clicked: {e.control.text}"))
            page.close(cupertino_alert_dialog)

        cupertino_alert_dialog = ft.CupertinoAlertDialog(
            title=ft.Text("Cupertino Alert Dialog"),
            content=ft.Text("Do you want to delete this file?"),
            on_dismiss=dialog_dismissed,
            actions=[
                ft.CupertinoDialogAction(
                    text="Yes",
                    is_destructive_action=True,
                    on_click=handle_action_click,
                ),
                ft.CupertinoDialogAction(
                    text="No",
                    is_default_action=True,
                    on_click=handle_action_click
                ),
            ],
        )

        page.add(
            ft.CupertinoFilledButton(
                text="Open CupertinoAlertDialog",
                on_click=lambda e: page.open(cupertino_alert_dialog),
            )
        )


    ft.app(target=main)
    ```

    -----

    Online docs: https://flet.dev/docs/controls/cupertinodialogaction
    """

    def __init__(
        self,
        text: Optional[str] = None,
        content: Optional[Control] = None,
        is_default_action: Optional[bool] = None,
        is_destructive_action: Optional[bool] = None,
        text_style: Optional[TextStyle] = None,
        on_click: OptionalControlEventCallable = None,
        #
        # Specific
        #
        ref: Optional[Ref] = None,
        opacity: OptionalNumber = None,
        visible: Optional[bool] = None,
        data: Any = None,
    ):
        Control.__init__(
            self,
            ref=ref,
            opacity=opacity,
            visible=visible,
            data=data,
        )

        self.text = text
        self.content = content
        self.on_click = on_click
        self.is_default_action = is_default_action
        self.is_destructive_action = is_destructive_action
        self.text_style = text_style

    def _get_control_name(self):
        return "cupertinodialogaction"

    def before_update(self):
        super().before_update()
        self._set_attr_json("textStyle", self.__text_style)

    def _get_children(self):
        if self.__content is None:
            return []
        self.__content._set_attr_internal("n", "content")
        return [self.__content]

    # text
    @property
    def text(self) -> Optional[str]:
        return self._get_attr("text")

    @text.setter
    def text(self, value: Optional[str]):
        self._set_attr("text", value)

    # is_default_action
    @property
    def is_default_action(self) -> bool:
        return self._get_attr("isDefaultAction", data_type="bool", def_value=False)

    @is_default_action.setter
    def is_default_action(self, value: Optional[bool]):
        self._set_attr("isDefaultAction", value)

    # is_destructive_action
    @property
    def is_destructive_action(self) -> bool:
        return self._get_attr("isDestructiveAction", data_type="bool", def_value=False)

    @is_destructive_action.setter
    def is_destructive_action(self, value: Optional[bool]):
        self._set_attr("isDestructiveAction", value)

    # on_click
    @property
    def on_click(self) -> OptionalControlEventCallable:
        return self._get_event_handler("click")

    @on_click.setter
    def on_click(self, handler: OptionalControlEventCallable):
        self._add_event_handler("click", handler)

    # content
    @property
    def content(self) -> Optional[Control]:
        return self.__content

    @content.setter
    def content(self, value: Optional[Control]):
        self.__content = value

    # text_style
    @property
    def text_style(self) -> Optional[TextStyle]:
        return self.__text_style

    @text_style.setter
    def text_style(self, value: Optional[TextStyle]):
        self.__text_style = value



from typing import Any, Optional

from flet.core.canvas.shape import Shape
from flet.core.control import OptionalNumber
from flet.core.painting import Paint


class Line(Shape):
    def __init__(
        self,
        x1: OptionalNumber = None,
        y1: OptionalNumber = None,
        x2: OptionalNumber = None,
        y2: OptionalNumber = None,
        paint: Optional[Paint] = None,
        #
        # Control
        #
        ref=None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        Shape.__init__(self, ref=ref, visible=visible, disabled=disabled, data=data)

        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.paint = paint

    def _get_control_name(self):
        return "line"

    def before_update(self):
        super().before_update()
        self._set_attr_json("paint", self.__paint)

    # x1
    @property
    def x1(self) -> OptionalNumber:
        return self._get_attr("x1")

    @x1.setter
    def x1(self, value: OptionalNumber):
        self._set_attr("x1", value)

    # y1
    @property
    def y1(self) -> OptionalNumber:
        return self._get_attr("y1")

    @y1.setter
    def y1(self, value: OptionalNumber):
        self._set_attr("y1", value)

    # x2
    @property
    def x2(self) -> OptionalNumber:
        return self._get_attr("x2")

    @x2.setter
    def x2(self, value: OptionalNumber):
        self._set_attr("x2", value)

    # y2
    @property
    def y2(self) -> OptionalNumber:
        return self._get_attr("y2")

    @y2.setter
    def y2(self, value: OptionalNumber):
        self._set_attr("y2", value)

    # paint
    @property
    def paint(self) -> Optional[Paint]:
        return self.__paint

    @paint.setter
    def paint(self, value: Optional[Paint]):
        self.__paint = value



from enum import Enum
from typing import Any, List, Optional

from flet.core.canvas.shape import Shape
from flet.core.painting import Paint
from flet.core.types import OffsetValue


class PointMode(Enum):
    POINTS = "points"
    LINES = "lines"
    POLYGON = "polygon"


class Points(Shape):
    def __init__(
        self,
        points: Optional[List[OffsetValue]] = None,
        point_mode: Optional[PointMode] = None,
        paint: Optional[Paint] = None,
        #
        # Control
        #
        ref=None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        Shape.__init__(self, ref=ref, visible=visible, disabled=disabled, data=data)

        self.points = points
        self.point_mode = point_mode
        self.paint = paint

    def _get_control_name(self):
        return "points"

    def before_update(self):
        super().before_update()
        self._set_attr_json("points", self.__points)
        self._set_attr_json("paint", self.__paint)

    # point_mode
    @property
    def point_mode(self) -> Optional[PointMode]:
        return self.__point_mode

    @point_mode.setter
    def point_mode(self, value: Optional[PointMode]):
        self.__point_mode = value
        self._set_attr(
            "pointMode", value.value if isinstance(value, PointMode) else value
        )

    # points
    @property
    def points(self) -> Optional[List[OffsetValue]]:
        return self.__points

    @points.setter
    def points(self, value: Optional[List[OffsetValue]]):
        self.__points = value if value is not None else []

    # paint
    @property
    def paint(self) -> Optional[Paint]:
        return self.__paint

    @paint.setter
    def paint(self, value: Optional[Paint]):
        self.__paint = value



from dataclasses import dataclass
from typing import List, Optional, Union

from flet.core.border import Border
from flet.core.box import BoxShadow, BoxShape, DecorationImage
from flet.core.gradients import Gradient
from flet.core.text_style import TextStyle
from flet.core.types import (
    BlendMode,
    BorderRadiusValue,
    ColorValue,
    DurationValue,
    MarginValue,
    OptionalNumber,
    PaddingValue,
    TextAlign,
)


class TooltipTriggerMode:
    MANUAL = "manual"
    TAP = "tap"
    LONG_PRESS = "long_press"


@dataclass
class Tooltip:
    """Tooltips provide text labels which help explain the function of a button or other user interface action."""

    message: str
    enable_feedback: Optional[bool] = None
    height: OptionalNumber = None
    vertical_offset: OptionalNumber = None
    margin: Optional[MarginValue] = None
    padding: Optional[PaddingValue] = None
    bgcolor: Optional[ColorValue] = None
    image: Optional[DecorationImage] = None
    shadow: Optional[List[BoxShadow]] = None
    blend_mode: Optional[BlendMode] = None
    gradient: Optional[Gradient] = None
    border: Optional[Border] = None
    border_radius: Optional[BorderRadiusValue] = None
    shape: Optional[BoxShape] = None
    text_style: Optional[TextStyle] = None
    text_align: Optional[TextAlign] = None
    prefer_below: Optional[bool] = None
    show_duration: Optional[DurationValue] = None
    wait_duration: Optional[DurationValue] = None
    exit_duration: Optional[DurationValue] = None
    enable_tap_to_dismiss: Optional[bool] = None
    exclude_from_semantics: Optional[bool] = None
    trigger_mode: Optional[TooltipTriggerMode] = None


TooltipValue = Union[str, "Tooltip"]



import dataclasses
import time
from enum import Enum
from typing import Any, List, Optional, Union

from flet.core.adaptive_control import AdaptiveControl
from flet.core.animation import AnimationValue
from flet.core.autofill_group import AutofillHint
from flet.core.badge import BadgeValue
from flet.core.box import BoxConstraints
from flet.core.control import Control, OptionalNumber
from flet.core.form_field_control import FormFieldControl, InputBorder
from flet.core.ref import Ref
from flet.core.text_style import StrutStyle, TextStyle
from flet.core.tooltip import TooltipValue
from flet.core.types import (
    BorderRadiusValue,
    Brightness,
    ClipBehavior,
    ColorEnums,
    ColorValue,
    DurationValue,
    IconValueOrControl,
    MouseCursor,
    OffsetValue,
    OptionalControlEventCallable,
    PaddingValue,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
    TextAlign,
    VerticalAlignment,
)

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class KeyboardType(Enum):
    NONE = "none"
    TEXT = "text"
    MULTILINE = "multiline"
    NUMBER = "number"
    PHONE = "phone"
    DATETIME = "datetime"
    EMAIL = "email"
    URL = "url"
    VISIBLE_PASSWORD = "visiblePassword"
    NAME = "name"
    STREET_ADDRESS = "streetAddress"


class TextCapitalization(Enum):
    CHARACTERS = "characters"
    WORDS = "words"
    SENTENCES = "sentences"


@dataclasses.dataclass
class InputFilter:
    regex_string: str
    allow: bool = True
    replacement_string: str = ""
    multiline: bool = False
    case_sensitive: bool = True
    unicode: bool = False
    dot_all: bool = False


class NumbersOnlyInputFilter(InputFilter):
    def __init__(self):
        super().__init__(regex_string=r"^[0-9]*$", allow=True, replacement_string="")


class TextOnlyInputFilter(InputFilter):
    def __init__(self):
        super().__init__(regex_string=r"^[a-zA-Z]*$", allow=True, replacement_string="")


class TextField(FormFieldControl, AdaptiveControl):
    """
    A text field lets the user enter text, either with hardware keyboard or with an onscreen keyboard.

    Example:
    ```
    import flet as ft

    def main(page: ft.Page):
        def button_clicked(e):
            t.value = f"Textboxes values are:  '{tb1.value}', '{tb2.value}', '{tb3.value}', '{tb4.value}', '{tb5.value}'."
            page.update()

        t = ft.Text()
        tb1 = ft.TextField(label="Standard")
        tb2 = ft.TextField(label="Disabled", disabled=True, value="First name")
        tb3 = ft.TextField(label="Read-only", read_only=True, value="Last name")
        tb4 = ft.TextField(label="With placeholder", hint_text="Please enter text here")
        tb5 = ft.TextField(label="With an icon", icon=ft.icons.EMOJI_EMOTIONS)
        b = ft.ElevatedButton(text="Submit", on_click=button_clicked)
        page.add(tb1, tb2, tb3, tb4, tb5, b, t)

    ft.app(target=main)
    ```

    -----

    Online docs: https://flet.dev/docs/controls/textfield
    """

    def __init__(
        self,
        value: Optional[str] = None,
        keyboard_type: Optional[KeyboardType] = None,
        multiline: Optional[bool] = None,
        min_lines: Optional[int] = None,
        max_lines: Optional[int] = None,
        max_length: Optional[int] = None,
        password: Optional[bool] = None,
        can_reveal_password: Optional[bool] = None,
        read_only: Optional[bool] = None,
        shift_enter: Optional[bool] = None,
        text_align: Optional[TextAlign] = None,
        autofocus: Optional[bool] = None,
        capitalization: Optional[TextCapitalization] = None,
        autocorrect: Optional[bool] = None,
        enable_suggestions: Optional[bool] = None,
        smart_dashes_type: Optional[bool] = None,
        smart_quotes_type: Optional[bool] = None,
        show_cursor: Optional[bool] = None,
        cursor_color: Optional[ColorValue] = None,
        cursor_error_color: Optional[ColorValue] = None,
        cursor_width: OptionalNumber = None,
        cursor_height: OptionalNumber = None,
        cursor_radius: OptionalNumber = None,
        selection_color: Optional[ColorValue] = None,
        input_filter: Optional[InputFilter] = None,
        obscuring_character: Optional[str] = None,
        enable_interactive_selection: Optional[bool] = None,
        enable_ime_personalized_learning: Optional[bool] = None,
        can_request_focus: Optional[bool] = None,
        ignore_pointers: Optional[bool] = None,
        enable_scribble: Optional[bool] = None,
        animate_cursor_opacity: Optional[bool] = None,
        always_call_on_tap: Optional[bool] = None,
        scroll_padding: Optional[PaddingValue] = None,
        clip_behavior: Optional[ClipBehavior] = None,
        keyboard_brightness: Optional[Brightness] = None,
        mouse_cursor: Optional[MouseCursor] = None,
        strut_style: Optional[StrutStyle] = None,
        autofill_hints: Union[None, AutofillHint, List[AutofillHint]] = None,
        on_change: OptionalControlEventCallable = None,
        on_click: OptionalControlEventCallable = None,
        on_submit: OptionalControlEventCallable = None,
        on_focus: OptionalControlEventCallable = None,
        on_blur: OptionalControlEventCallable = None,
        on_tap_outside: OptionalControlEventCallable = None,
        #
        # FormField
        #
        text_size: OptionalNumber = None,
        text_style: Optional[TextStyle] = None,
        text_vertical_align: Union[VerticalAlignment, OptionalNumber] = None,
        label: Optional[Union[str, Control]] = None,
        label_style: Optional[TextStyle] = None,
        icon: Optional[IconValueOrControl] = None,
        border: Optional[InputBorder] = None,
        color: Optional[ColorValue] = None,
        bgcolor: Optional[ColorValue] = None,
        border_radius: Optional[BorderRadiusValue] = None,
        border_width: OptionalNumber = None,
        border_color: Optional[ColorValue] = None,
        focused_color: Optional[ColorValue] = None,
        focused_bgcolor: Optional[ColorValue] = None,
        focused_border_width: OptionalNumber = None,
        focused_border_color: Optional[ColorValue] = None,
        content_padding: Optional[PaddingValue] = None,
        dense: Optional[bool] = None,
        filled: Optional[bool] = None,
        fill_color: Optional[ColorValue] = None,
        hover_color: Optional[ColorValue] = None,
        hint_text: Optional[str] = None,
        hint_style: Optional[TextStyle] = None,
        helper: Optional[Control] = None,
        helper_text: Optional[str] = None,
        helper_style: Optional[TextStyle] = None,
        counter: Optional[Control] = None,
        counter_text: Optional[str] = None,
        counter_style: Optional[TextStyle] = None,
        error: Optional[Control] = None,
        error_text: Optional[str] = None,
        error_style: Optional[TextStyle] = None,
        prefix: Optional[Control] = None,
        prefix_icon: Optional[IconValueOrControl] = None,
        prefix_text: Optional[str] = None,
        prefix_style: Optional[TextStyle] = None,
        suffix: Optional[Control] = None,
        suffix_icon: Optional[IconValueOrControl] = None,
        suffix_text: Optional[str] = None,
        suffix_style: Optional[TextStyle] = None,
        focus_color: Optional[ColorValue] = None,
        align_label_with_hint: Optional[bool] = None,
        hint_fade_duration: Optional[DurationValue] = None,
        hint_max_lines: Optional[int] = None,
        helper_max_lines: Optional[int] = None,
        error_max_lines: Optional[int] = None,
        prefix_icon_size_constraints: Optional[BoxConstraints] = None,
        suffix_icon_size_constraints: Optional[BoxConstraints] = None,
        size_constraints: Optional[BoxConstraints] = None,
        collapsed: Optional[bool] = None,
        fit_parent_size: Optional[bool] = None,
        #
        # ConstrainedControl and AdaptiveControl
        #
        ref: Optional[Ref] = None,
        key: Optional[str] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        tooltip: Optional[TooltipValue] = None,
        badge: Optional[BadgeValue] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
        rtl: Optional[bool] = None,
        adaptive: Optional[bool] = None,
    ):
        FormFieldControl.__init__(
            self,
            ref=ref,
            key=key,
            width=width,
            height=height,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            tooltip=tooltip,
            badge=badge,
            visible=visible,
            disabled=disabled,
            data=data,
            rtl=rtl,
            #
            # FormField
            #
            text_size=text_size,
            text_style=text_style,
            text_vertical_align=text_vertical_align,
            label=label,
            label_style=label_style,
            icon=icon,
            border=border,
            color=color,
            bgcolor=bgcolor,
            border_radius=border_radius,
            border_width=border_width,
            border_color=border_color,
            focused_color=focused_color,
            focused_bgcolor=focused_bgcolor,
            focused_border_width=focused_border_width,
            focused_border_color=focused_border_color,
            content_padding=content_padding,
            dense=dense,
            filled=filled,
            fill_color=fill_color,
            hover_color=hover_color,
            hint_text=hint_text,
            hint_style=hint_style,
            helper=helper,
            helper_text=helper_text,
            helper_style=helper_style,
            counter=counter,
            counter_text=counter_text,
            counter_style=counter_style,
            error=error,
            error_text=error_text,
            error_style=error_style,
            prefix=prefix,
            prefix_icon=prefix_icon,
            prefix_text=prefix_text,
            prefix_style=prefix_style,
            suffix=suffix,
            suffix_icon=suffix_icon,
            suffix_text=suffix_text,
            suffix_style=suffix_style,
            focus_color=focus_color,
            align_label_with_hint=align_label_with_hint,
            hint_fade_duration=hint_fade_duration,
            hint_max_lines=hint_max_lines,
            helper_max_lines=helper_max_lines,
            error_max_lines=error_max_lines,
            prefix_icon_size_constraints=prefix_icon_size_constraints,
            suffix_icon_size_constraints=suffix_icon_size_constraints,
            size_constraints=size_constraints,
            collapsed=collapsed,
            fit_parent_size=fit_parent_size,
        )

        AdaptiveControl.__init__(self, adaptive=adaptive)

        self.value = value
        self.text_style = text_style
        self.keyboard_type = keyboard_type
        self.text_align = text_align
        self.multiline = multiline
        self.min_lines = min_lines
        self.max_lines = max_lines
        self.max_length = max_length
        self.read_only = read_only
        self.shift_enter = shift_enter
        self.password = password
        self.can_reveal_password = can_reveal_password
        self.autofocus = autofocus
        self.capitalization = capitalization
        self.autocorrect = autocorrect
        self.show_cursor = show_cursor
        self.enable_suggestions = enable_suggestions
        self.smart_dashes_type = smart_dashes_type
        self.smart_quotes_type = smart_quotes_type
        self.cursor_color = cursor_color
        self.cursor_height = cursor_height
        self.cursor_width = cursor_width
        self.cursor_radius = cursor_radius
        self.selection_color = selection_color
        self.input_filter = input_filter
        self.autofill_hints = autofill_hints
        self.on_change = on_change
        self.on_submit = on_submit
        self.on_focus = on_focus
        self.on_blur = on_blur
        self.on_click = on_click
        self.obscuring_character = obscuring_character
        self.enable_scribble = enable_scribble
        self.strut_style = strut_style
        self.scroll_padding = scroll_padding
        self.cursor_error_color = cursor_error_color
        self.keyboard_brightness = keyboard_brightness
        self.mouse_cursor = mouse_cursor
        self.enable_interactive_selection = enable_interactive_selection
        self.enable_ime_personalized_learning = enable_ime_personalized_learning
        self.can_request_focus = can_request_focus
        self.ignore_pointers = ignore_pointers
        self.animate_cursor_opacity = animate_cursor_opacity
        self.always_call_on_tap = always_call_on_tap
        self.clip_behavior = clip_behavior
        self.on_tap_outside = on_tap_outside

    def _get_control_name(self):
        return "textfield"

    def before_update(self):
        super().before_update()
        assert (
            self.max_lines is None
            or self.min_lines is None
            or self.min_lines <= self.max_lines
        ), "min_lines can't be greater than max_lines"
        self._set_attr_json("inputFilter", self.__input_filter)
        self._set_attr_json("autofillHints", self.__autofill_hints)
        self._set_attr_json("scrollPadding", self.__scroll_padding)
        self._set_attr_json("strutStyle", self.__strut_style)
        if (
            (
                self.bgcolor is not None
                or self.fill_color is not None
                or self.hover_color is not None
                or self.focused_color is not None
            )
        ) and self.filled is None:
            self.filled = True  # required to display any of the above colors

    def focus(self):
        self._set_attr_json("focus", str(time.time()))
        self.update()

    def blur(self):
        self._set_attr_json("blur", str(time.time()))
        self.update()

    # value
    @property
    def value(self) -> Optional[str]:
        return self._get_attr("value", def_value="")

    @value.setter
    def value(self, value: Optional[str]):
        self._set_attr("value", value)

    # strut_style
    @property
    def strut_style(self) -> Optional[TextStyle]:
        return self.__strut_style

    @strut_style.setter
    def strut_style(self, value: Optional[TextStyle]):
        self.__strut_style = value

    # cursor_error_color
    @property
    def cursor_error_color(self) -> Optional[str]:
        return self._get_attr("cursorErrorColor")

    @cursor_error_color.setter
    def cursor_error_color(self, value: Optional[str]):
        self._set_attr("cursorErrorColor", value)

    # enable_interactive_selection
    @property
    def enable_interactive_selection(self) -> bool:
        return self._get_attr(
            "enableInteractiveSelection", data_type="bool", def_value=True
        )

    @enable_interactive_selection.setter
    def enable_interactive_selection(self, value: Optional[bool]):
        self._set_attr("enableInteractiveSelection", value)

    # enable_ime_personalized_learning
    @property
    def enable_ime_personalized_learning(self) -> bool:
        return self._get_attr(
            "enableIMEPersonalizedLearning", data_type="bool", def_value=True
        )

    @enable_ime_personalized_learning.setter
    def enable_ime_personalized_learning(self, value: Optional[bool]):
        self._set_attr("enableIMEPersonalizedLearning", value)

    # animate_cursor_opacity
    @property
    def animate_cursor_opacity(self) -> Optional[bool]:
        return self._get_attr("animateCursorOpacity", data_type="bool")

    @animate_cursor_opacity.setter
    def animate_cursor_opacity(self, value: Optional[bool]):
        self._set_attr("animateCursorOpacity", value)

    # keyboard_type
    @property
    def keyboard_type(self) -> Optional[KeyboardType]:
        return self.__keyboard_type

    @keyboard_type.setter
    def keyboard_type(self, value: Optional[KeyboardType]):
        self.__keyboard_type = value
        self._set_enum_attr("keyboardType", value, KeyboardType)

    # text_align
    @property
    def text_align(self) -> Optional[TextAlign]:
        return self.__text_align

    @text_align.setter
    def text_align(self, value: Optional[TextAlign]):
        self.__text_align = value
        self._set_enum_attr("textAlign", value, TextAlign)

    # multiline
    @property
    def multiline(self) -> bool:
        return self._get_attr("multiline", data_type="bool", def_value=False)

    @multiline.setter
    def multiline(self, value: Optional[bool]):
        self._set_attr("multiline", value)

    # min_lines
    @property
    def min_lines(self) -> Optional[int]:
        return self._get_attr("minLines")

    @min_lines.setter
    def min_lines(self, value: Optional[int]):
        assert value is None or value > 0, "min_lines must be greater than 0"
        self._set_attr("minLines", value)

    # max_lines
    @property
    def max_lines(self) -> Optional[int]:
        return self._get_attr("maxLines")

    @max_lines.setter
    def max_lines(self, value: Optional[int]):
        assert value is None or value > 0, "max_lines must be greater than 0"
        self._set_attr("maxLines", value)

    # max_length
    @property
    def max_length(self) -> Optional[int]:
        return self._get_attr("maxLength")

    @max_length.setter
    def max_length(self, value: Optional[int]):
        assert (
            value is None or value == -1 or value > 0
        ), "max_length must be either equal to -1 or greater than 0"
        self._set_attr("maxLength", value)

    # obscuring_character
    @property
    def obscuring_character(self) -> Optional[str]:
        return self._get_attr("obscuringCharacter", def_value="•")

    @obscuring_character.setter
    def obscuring_character(self, value: Optional[str]):
        self._set_attr("obscuringCharacter", value)

    # enable_scribble
    @property
    def enable_scribble(self) -> bool:
        return self._get_attr("enableScribble", data_type="bool", def_value=True)

    @enable_scribble.setter
    def enable_scribble(self, value: Optional[bool]):
        self._set_attr("enableScribble", value)

    # scroll_padding
    @property
    def scroll_padding(self) -> Optional[PaddingValue]:
        return self.__scroll_padding

    @scroll_padding.setter
    def scroll_padding(self, value: Optional[PaddingValue]):
        self.__scroll_padding = value

    # keyboard_brightness
    @property
    def keyboard_brightness(self) -> Optional[Brightness]:
        return self.__keyboard_brightness

    @keyboard_brightness.setter
    def keyboard_brightness(self, value: Optional[Brightness]):
        self.__keyboard_brightness = value
        self._set_enum_attr("keyboardBrightness", value, Brightness)

    # mouse_cursor
    @property
    def mouse_cursor(self) -> Optional[MouseCursor]:
        return self.__mouse_cursor

    @mouse_cursor.setter
    def mouse_cursor(self, value: Optional[MouseCursor]):
        self.__mouse_cursor = value
        self._set_enum_attr("mouseCursor", value, MouseCursor)

    # ignore_pointers
    @property
    def ignore_pointers(self) -> bool:
        return self._get_attr("ignorePointers", data_type="bool", def_value=False)

    @ignore_pointers.setter
    def ignore_pointers(self, value: Optional[bool]):
        self._set_attr("ignorePointers", value)

    # clip_behavior
    @property
    def clip_behavior(self) -> Optional[ClipBehavior]:
        return self.__clip_behavior

    @clip_behavior.setter
    def clip_behavior(self, value: Optional[ClipBehavior]):
        self.__clip_behavior = value
        self._set_enum_attr("clipBehavior", value, ClipBehavior)

    # can_request_focus
    @property
    def can_request_focus(self) -> bool:
        return self._get_attr("canRequestFocus", data_type="bool", def_value=True)

    @can_request_focus.setter
    def can_request_focus(self, value: Optional[bool]):
        self._set_attr("canRequestFocus", value)

    # always_call_on_tap
    @property
    def always_call_on_tap(self) -> bool:
        return self._get_attr("alwaysCallOnTap", data_type="bool", def_value=False)

    @always_call_on_tap.setter
    def always_call_on_tap(self, value: Optional[bool]):
        self._set_attr("alwaysCallOnTap", value)

    # read_only
    @property
    def read_only(self) -> bool:
        return self._get_attr("readOnly", data_type="bool", def_value=False)

    @read_only.setter
    def read_only(self, value: Optional[bool]):
        self._set_attr("readOnly", value)

    # shift_enter
    @property
    def shift_enter(self) -> bool:
        return self._get_attr("shiftEnter", data_type="bool", def_value=False)

    @shift_enter.setter
    def shift_enter(self, value: Optional[bool]):
        self._set_attr("shiftEnter", value)

    # password
    @property
    def password(self) -> bool:
        return self._get_attr("password", data_type="bool", def_value=False)

    @password.setter
    def password(self, value: Optional[bool]):
        self._set_attr("password", value)

    # can_reveal_password
    @property
    def can_reveal_password(self) -> bool:
        return self._get_attr("canRevealPassword", data_type="bool", def_value=False)

    @can_reveal_password.setter
    def can_reveal_password(self, value: Optional[bool]):
        self._set_attr("canRevealPassword", value)

    # autofocus
    @property
    def autofocus(self) -> bool:
        return self._get_attr("autofocus", data_type="bool", def_value=False)

    @autofocus.setter
    def autofocus(self, value: Optional[bool]):
        self._set_attr("autofocus", value)

    # capitalization
    @property
    def capitalization(self) -> Optional[TextCapitalization]:
        return self.__capitalization

    @capitalization.setter
    def capitalization(self, value: Optional[TextCapitalization]):
        self.__capitalization = value
        self._set_enum_attr("capitalization", value, TextCapitalization)

    # autocorrect
    @property
    def autocorrect(self) -> bool:
        return self._get_attr("autocorrect", data_type="bool", def_value=True)

    @autocorrect.setter
    def autocorrect(self, value: Optional[bool]):
        self._set_attr("autocorrect", value)

    # show_cursor
    @property
    def show_cursor(self) -> bool:
        return self._get_attr("showCursor", data_type="bool", def_value=True)

    @show_cursor.setter
    def show_cursor(self, value: Optional[bool]):
        self._set_attr("showCursor", value)

    # enable_suggestions
    @property
    def enable_suggestions(self) -> bool:
        return self._get_attr("enableSuggestions", data_type="bool", def_value=True)

    @enable_suggestions.setter
    def enable_suggestions(self, value: Optional[bool]):
        self._set_attr("enableSuggestions", value)

    # smart_dashes_type
    @property
    def smart_dashes_type(self) -> bool:
        return self._get_attr("smartDashesType", data_type="bool", def_value=True)

    @smart_dashes_type.setter
    def smart_dashes_type(self, value: Optional[bool]):
        self._set_attr("smartDashesType", value)

    # smart_quotes_type
    @property
    def smart_quotes_type(self) -> bool:
        return self._get_attr("smartQuotesType", data_type="bool", def_value=True)

    @smart_quotes_type.setter
    def smart_quotes_type(self, value: Optional[bool]):
        self._set_attr("smartQuotesType", value)

    # cursor_color
    @property
    def cursor_color(self):
        return self.__cursor_color

    @cursor_color.setter
    def cursor_color(self, value):
        self.__cursor_color = value
        self._set_enum_attr("cursorColor", value, ColorEnums)

    # cursor_height
    @property
    def cursor_height(self) -> OptionalNumber:
        return self._get_attr("cursorHeight")

    @cursor_height.setter
    def cursor_height(self, value: OptionalNumber):
        self._set_attr("cursorHeight", value)

    # cursor_width
    @property
    def cursor_width(self) -> OptionalNumber:
        return self._get_attr("cursorWidth")

    @cursor_width.setter
    def cursor_width(self, value: OptionalNumber):
        self._set_attr("cursorWidth", value)

    # cursor_radius
    @property
    def cursor_radius(self) -> OptionalNumber:
        return self._get_attr("cursorRadius")

    @cursor_radius.setter
    def cursor_radius(self, value: OptionalNumber):
        self._set_attr("cursorRadius", value)

    # selection_color
    @property
    def selection_color(self) -> Optional[ColorValue]:
        return self.__selection_color

    @selection_color.setter
    def selection_color(self, value: Optional[ColorValue]):
        self.__selection_color = value
        self._set_enum_attr("selectionColor", value, ColorEnums)

    # input_filter
    @property
    def input_filter(self) -> Optional[InputFilter]:
        return self.__input_filter

    @input_filter.setter
    def input_filter(self, value: Optional[InputFilter]):
        self.__input_filter = value

    # autofill_hints
    @property
    def autofill_hints(self) -> Union[None, AutofillHint, List[AutofillHint]]:
        return self.__autofill_hints

    @autofill_hints.setter
    def autofill_hints(self, value: Union[None, AutofillHint, List[AutofillHint]]):
        if value is not None:
            if isinstance(value, List):
                value = list(
                    map(
                        lambda x: x.value if isinstance(x, AutofillHint) else str(x),
                        value,
                    )
                )
            elif isinstance(value, AutofillHint):
                value = value.value
        self.__autofill_hints = value

    # on_change
    @property
    def on_change(self):
        return self._get_event_handler("change")

    @on_change.setter
    def on_change(self, handler: OptionalControlEventCallable):
        self._add_event_handler("change", handler)
        self._set_attr("onChange", True if handler is not None else None)

    # on_submit
    @property
    def on_submit(self) -> OptionalControlEventCallable:
        return self._get_event_handler("submit")

    @on_submit.setter
    def on_submit(self, handler: OptionalControlEventCallable):
        self._add_event_handler("submit", handler)

    # on_focus
    @property
    def on_focus(self) -> OptionalControlEventCallable:
        return self._get_event_handler("focus")

    @on_focus.setter
    def on_focus(self, handler: OptionalControlEventCallable):
        self._add_event_handler("focus", handler)

    # on_blur
    @property
    def on_blur(self) -> OptionalControlEventCallable:
        return self._get_event_handler("blur")

    @on_blur.setter
    def on_blur(self, handler: OptionalControlEventCallable):
        self._add_event_handler("blur", handler)

    # on_click
    @property
    def on_click(self):
        return self._get_event_handler("click")

    @on_click.setter
    def on_click(self, handler: OptionalControlEventCallable):
        self._add_event_handler("click", handler)

    # on_tap_outside
    @property
    def on_tap_outside(self) -> OptionalControlEventCallable:
        return self._get_event_handler("tapOutside")

    @on_tap_outside.setter
    def on_tap_outside(self, handler: OptionalControlEventCallable):
        self._add_event_handler("tapOutside", handler)
        self._set_attr("onTapOutside", True if handler is not None else None)



from enum import Enum
from typing import Any, Callable, List, Optional, Union

from flet.core.animation import AnimationValue
from flet.core.buttons import OutlinedBorder
from flet.core.constrained_control import ConstrainedControl
from flet.core.control import Control
from flet.core.ref import Ref
from flet.core.text_style import TextStyle
from flet.core.types import (
    ColorEnums,
    ColorValue,
    IconEnums,
    IconValueOrControl,
    OffsetValue,
    OptionalControlEventCallable,
    OptionalNumber,
    PaddingValue,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
)


class NavigationRailLabelType(Enum):
    NONE = "none"
    ALL = "all"
    SELECTED = "selected"


class NavigationRailDestination(Control):
    def __init__(
        self,
        icon: Optional[IconValueOrControl] = None,
        selected_icon: Optional[IconValueOrControl] = None,
        label: Optional[str] = None,
        label_content: Optional[Control] = None,
        padding: Optional[PaddingValue] = None,
        indicator_color: Optional[ColorValue] = None,
        indicator_shape: Optional[OutlinedBorder] = None,
        #
        # Control
        #
        ref: Optional[Ref] = None,
        disabled: Optional[bool] = None,
        visible: Optional[bool] = None,
        data: Any = None,
    ) -> None:
        Control.__init__(self, ref=ref, disabled=disabled, visible=visible, data=data)
        self.label = label
        self.icon = icon
        self.selected_icon = selected_icon
        self.label_content = label_content
        self.padding = padding
        self.indicator_color = indicator_color
        self.indicator_shape = indicator_shape

    def _get_control_name(self):
        return "navigationraildestination"

    def before_update(self) -> None:
        super().before_update()
        self._set_attr_json("padding", self.__padding)
        if isinstance(self.__indicator_shape, OutlinedBorder):
            self._set_attr_json("indicatorShape", self.__indicator_shape)

    def _get_children(self):
        children = []
        if self.__label_content:
            self.__label_content._set_attr_internal("n", "label_content")
            children.append(self.__label_content)
        if isinstance(self.__icon, Control):
            self.__icon._set_attr_internal("n", "icon")
            children.append(self.__icon)
        if isinstance(self.__selected_icon, Control):
            self.__selected_icon._set_attr_internal("n", "selected_icon")
            children.append(self.__selected_icon)
        return children

    # icon
    @property
    def icon(self) -> Optional[IconValueOrControl]:
        return self.__icon

    @icon.setter
    def icon(self, value: Optional[IconValueOrControl]):
        self.__icon = value
        if not isinstance(value, Control):
            self._set_enum_attr("icon", value, IconEnums)

    # selected_icon
    @property
    def selected_icon(self) -> Optional[IconValueOrControl]:
        return self.__selected_icon

    @selected_icon.setter
    def selected_icon(self, value: Optional[IconValueOrControl]):
        self.__selected_icon = value
        if not isinstance(value, Control):
            self._set_enum_attr("selectedIcon", value, IconEnums)

    # label
    @property
    def label(self) -> Optional[str]:
        return self._get_attr("label")

    @label.setter
    def label(self, value: Optional[str]):
        self._set_attr("label", value)

    # label_content
    @property
    def label_content(self) -> Optional[Control]:
        return self.__label_content

    @label_content.setter
    def label_content(self, value: Optional[Control]):
        self.__label_content = value

    # indicator_color
    @property
    def indicator_color(self) -> Optional[ColorValue]:
        return self.__indicator_color

    @indicator_color.setter
    def indicator_color(self, value: Optional[ColorValue]):
        self.__indicator_color = value
        self._set_enum_attr("indicatorColor", value, ColorEnums)

    # indicator_shape
    @property
    def indicator_shape(self) -> Optional[OutlinedBorder]:
        return self.__indicator_shape

    @indicator_shape.setter
    def indicator_shape(self, value: Optional[OutlinedBorder]):
        self.__indicator_shape = value

    # padding
    @property
    def padding(self) -> Optional[PaddingValue]:
        return self.__padding

    @padding.setter
    def padding(self, value: Optional[PaddingValue]):
        self.__padding = value


class NavigationRail(ConstrainedControl):
    """
    A material widget that is meant to be displayed at the left or right of an app to navigate between a small number of views, typically between three and five.

    Example:

    ```
    import flet as ft

    def main(page: ft.Page):

        rail = ft.NavigationRail(
            selected_index=1,
            label_type=ft.NavigationRailLabelType.ALL,
            # extended=True,
            min_width=100,
            min_extended_width=400,
            leading=ft.FloatingActionButton(icon=ft.icons.CREATE, text="Add"),
            group_alignment=-0.9,
            destinations=[
                ft.NavigationRailDestination(
                    icon=ft.icons.FAVORITE_BORDER, selected_icon=ft.icons.FAVORITE, label="First"
                ),
                ft.NavigationRailDestination(
                    icon_content=ft.Icon(ft.icons.BOOKMARK_BORDER),
                    selected_icon_content=ft.Icon(ft.icons.BOOKMARK),
                    label="Second",
                ),
                ft.NavigationRailDestination(
                    icon=ft.icons.SETTINGS_OUTLINED,
                    selected_icon_content=ft.Icon(ft.icons.SETTINGS),
                    label_content=ft.Text("Settings"),
                ),
            ],
            on_change=lambda e: print("Selected destination:", e.control.selected_index),
        )

        page.add(
            ft.Row(
                [
                    rail,
                    ft.VerticalDivider(width=1),
                    ft.Column([ ft.Text("Body!")], alignment=ft.MainAxisAlignment.START, expand=True),
                ],
                expand=True,
            )
        )

    ft.app(target=main)
    ```

    -----

    Online docs: https://flet.dev/docs/controls/navigationrail
    """

    def __init__(
        self,
        destinations: Optional[List[NavigationRailDestination]] = None,
        elevation: OptionalNumber = None,
        selected_index: Optional[int] = None,
        extended: Optional[bool] = None,
        label_type: Optional[NavigationRailLabelType] = None,
        bgcolor: Optional[ColorValue] = None,
        indicator_color: Optional[ColorValue] = None,
        indicator_shape: Optional[OutlinedBorder] = None,
        leading: Optional[Control] = None,
        trailing: Optional[Control] = None,
        min_width: OptionalNumber = None,
        min_extended_width: OptionalNumber = None,
        group_alignment: OptionalNumber = None,
        selected_label_text_style: Optional[TextStyle] = None,
        unselected_label_text_style: Optional[TextStyle] = None,
        on_change: OptionalControlEventCallable = None,
        #
        # ConstrainedControl
        #
        ref: Optional[Ref] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: Callable[..., None] = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
        rtl: Optional[bool] = False,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            visible=visible,
            disabled=disabled,
            data=data,
            rtl=rtl,
        )

        self.destinations = destinations
        self.selected_index = selected_index
        self.elevation = elevation
        self.extended = extended
        self.label_type = label_type
        self.bgcolor = bgcolor
        self.indicator_color = indicator_color
        self.indicator_shape = indicator_shape
        self.leading = leading
        self.trailing = trailing
        self.min_width = min_width
        self.min_extended_width = min_extended_width
        self.group_alignment = group_alignment
        self.on_change = on_change
        self.selected_label_text_style = selected_label_text_style
        self.unselected_label_text_style = unselected_label_text_style

    def _get_control_name(self):
        return "navigationrail"

    def before_update(self):
        super().before_update()
        if isinstance(self.__indicator_shape, OutlinedBorder):
            self._set_attr_json("indicatorShape", self.__indicator_shape)
        if isinstance(self.__selected_label_text_style, TextStyle):
            self._set_attr_json(
                "selectedLabelTextStyle", self.__selected_label_text_style
            )
        if isinstance(self.__unselected_label_text_style, TextStyle):
            self._set_attr_json(
                "unselectedLabelTextStyle", self.__unselected_label_text_style
            )

    def _get_children(self):
        children = []
        if self.__leading:
            self.__leading._set_attr_internal("n", "leading")
            children.append(self.__leading)
        if self.__trailing:
            self.__trailing._set_attr_internal("n", "trailing")
            children.append(self.__trailing)
        return children + self.__destinations

    # destinations
    @property
    def destinations(self) -> Optional[List[NavigationRailDestination]]:
        return self.__destinations

    @destinations.setter
    def destinations(self, value: Optional[List[NavigationRailDestination]]):
        self.__destinations = value if value else []

    # on_change
    @property
    def on_change(self) -> OptionalControlEventCallable:
        return self._get_event_handler("change")

    @on_change.setter
    def on_change(self, handler: OptionalControlEventCallable):
        self._add_event_handler("change", handler)

    # selected_index
    @property
    def selected_index(self) -> int:
        return self._get_attr("selectedIndex", data_type="int", def_value=0)

    @selected_index.setter
    def selected_index(self, value: Optional[int]):
        self._set_attr("selectedIndex", value)

    # label_type
    @property
    def label_type(self) -> Optional[NavigationRailLabelType]:
        return self.__label_type

    @label_type.setter
    def label_type(self, value: Optional[NavigationRailLabelType]):
        self.__label_type = value
        self._set_enum_attr("labelType", value, NavigationRailLabelType)

    # indicator_shape
    @property
    def indicator_shape(self) -> Optional[OutlinedBorder]:
        return self.__indicator_shape

    @indicator_shape.setter
    def indicator_shape(self, value: Optional[OutlinedBorder]):
        self.__indicator_shape = value

    # indicator_color
    @property
    def indicator_color(self) -> Optional[ColorValue]:
        return self.__indicator_color

    @indicator_color.setter
    def indicator_color(self, value: Optional[ColorValue]):
        self.__indicator_color = value
        self._set_enum_attr("indicatorColor", value, ColorEnums)

    # bgcolor
    @property
    def bgcolor(self) -> Optional[ColorValue]:
        return self.__bgcolor

    @bgcolor.setter
    def bgcolor(self, value: Optional[ColorValue]):
        self.__bgcolor = value
        self._set_enum_attr("bgcolor", value, ColorEnums)

    # elevation
    @property
    def elevation(self) -> OptionalNumber:
        return self._get_attr("elevation")

    @elevation.setter
    def elevation(self, value: OptionalNumber):
        assert value is None or value > 0, "elevation must be greater than 0"
        self._set_attr("elevation", value)

    # extended
    @property
    def extended(self) -> bool:
        return self._get_attr("extended", data_type="bool", def_value=False)

    @extended.setter
    def extended(self, value: Optional[bool]):
        self._set_attr("extended", value)

    # leading
    @property
    def leading(self) -> Optional[Control]:
        return self.__leading

    @leading.setter
    def leading(self, value: Optional[Control]):
        self.__leading = value

    # trailing
    @property
    def trailing(self) -> Optional[Control]:
        return self.__trailing

    @trailing.setter
    def trailing(self, value: Optional[Control]):
        self.__trailing = value

    # selected_label_text_style
    @property
    def selected_label_text_style(self) -> Optional[TextStyle]:
        return self.__selected_label_text_style

    @selected_label_text_style.setter
    def selected_label_text_style(self, value: Optional[TextStyle]):
        self.__selected_label_text_style = value

    # unselected_label_text_style
    @property
    def unselected_label_text_style(self) -> Optional[TextStyle]:
        return self.__unselected_label_text_style

    @unselected_label_text_style.setter
    def unselected_label_text_style(self, value: Optional[TextStyle]):
        self.__unselected_label_text_style = value

    # min_width
    @property
    def min_width(self) -> OptionalNumber:
        return self._get_attr("minWidth", data_type="float")

    @min_width.setter
    def min_width(self, value: OptionalNumber):
        self._set_attr("minWidth", value)

    # min_extended_width
    @property
    def min_extended_width(self) -> OptionalNumber:
        return self._get_attr("minExtendedWidth", data_type="float")

    @min_extended_width.setter
    def min_extended_width(self, value: OptionalNumber):
        self._set_attr("minExtendedWidth", value)

    # group_alignment
    @property
    def group_alignment(self) -> OptionalNumber:
        return self._get_attr("groupAlignment", data_type="float")

    @group_alignment.setter
    def group_alignment(self, value: OptionalNumber):
        self._set_attr("groupAlignment", value)



from typing import Any, List, Optional, Union

from flet.core.adaptive_control import AdaptiveControl
from flet.core.animation import AnimationValue
from flet.core.border import BorderSide
from flet.core.constrained_control import ConstrainedControl
from flet.core.control import Control, OptionalNumber
from flet.core.form_field_control import IconValueOrControl
from flet.core.ref import Ref
from flet.core.text_style import TextStyle
from flet.core.types import (
    BorderRadiusValue,
    ClipBehavior,
    ColorEnums,
    ColorValue,
    ControlStateValue,
    IconEnums,
    IconValue,
    MarginValue,
    MouseCursor,
    OffsetValue,
    OptionalControlEventCallable,
    PaddingValue,
    ResponsiveNumber,
    RotateValue,
    ScaleValue,
    TabAlignment,
)


class Tab(AdaptiveControl):
    def __init__(
        self,
        text: Optional[str] = None,
        content: Optional[Control] = None,
        tab_content: Optional[Control] = None,
        icon: Optional[IconValueOrControl] = None,
        height: OptionalNumber = None,
        icon_margin: Optional[MarginValue] = None,
        #
        # Control and AdaptiveControl
        #
        ref: Optional[Ref] = None,
        visible: Optional[bool] = None,
        adaptive: Optional[bool] = None,
    ):
        Control.__init__(self, ref=ref, visible=visible)
        AdaptiveControl.__init__(self, adaptive=adaptive)

        self.text = text
        self.icon = icon
        self.content = content
        self.tab_content = tab_content
        self.height = height
        self.icon_margin = icon_margin

    def _get_control_name(self):
        return "tab"

    def _get_children(self):
        children = []
        if self.__tab_content:
            self.__tab_content._set_attr_internal("n", "tab_content")
            children.append(self.__tab_content)
        if self.__content:
            self.__content._set_attr_internal("n", "content")
            children.append(self.__content)
        if isinstance(self.__icon, Control):
            self.__icon._set_attr_internal("n", "icon")
            children.append(self.__icon)
        return children

    def before_update(self):
        super().before_update()
        self._set_attr_json("iconMargin", self.__icon_margin)
        if isinstance(self.__icon, IconValue):
            self._set_enum_attr("icon", self.__icon, IconEnums)

    # text
    @property
    def text(self) -> Optional[str]:
        return self._get_attr("text")

    @text.setter
    def text(self, value: Optional[str]):
        self._set_attr("text", value)

    # height
    @property
    def height(self) -> Optional[float]:
        return self._get_attr("height", data_type="float")

    @height.setter
    def height(self, value: OptionalNumber):
        self._set_attr("height", value)

    # icon
    @property
    def icon(self) -> Optional[IconValueOrControl]:
        return self.__icon

    @icon.setter
    def icon(self, value: Optional[IconValueOrControl]):
        self.__icon = value

    # tab_content
    @property
    def tab_content(self) -> Optional[Control]:
        return self.__tab_content

    @tab_content.setter
    def tab_content(self, value: Optional[Control]):
        self.__tab_content = value

    # content
    @property
    def content(self) -> Optional[Control]:
        return self.__content

    @content.setter
    def content(self, value: Optional[Control]):
        self.__content = value

    # icon_margin
    @property
    def icon_margin(self) -> Optional[MarginValue]:
        return self.__icon_margin

    @icon_margin.setter
    def icon_margin(self, value: Optional[MarginValue]):
        self.__icon_margin = value


class Tabs(ConstrainedControl, AdaptiveControl):
    """
    The Tabs control is used for navigating frequently accessed, distinct content categories. Tabs allow for navigation between two or more content views and relies on text headers to articulate the different sections of content.

    Example:
    ```
    import flet as ft


    def main(page: ft.Page):

        t = ft.Tabs(
            selected_index=1,
            animation_duration=300,
            tabs=[
                ft.Tab(
                    text="Tab 1",
                    content=ft.Container(
                        content=ft.Text("This is Tab 1"), alignment=ft.alignment.center
                    ),
                ),
                ft.Tab(
                    tab_content=ft.Icon(ft.icons.SEARCH),
                    content=ft.Text("This is Tab 2"),
                ),
                ft.Tab(
                    text="Tab 3",
                    icon=ft.icons.SETTINGS,
                    content=ft.Text("This is Tab 3"),
                ),
            ],
            expand=1,
        )

        page.add(t)


    ft.app(target=main)

    ```

    -----

    Online docs: https://flet.dev/docs/controls/tabs
    """

    def __init__(
        self,
        tabs: Optional[List[Tab]] = None,
        selected_index: Optional[int] = None,
        scrollable: Optional[bool] = None,
        tab_alignment: Optional[TabAlignment] = None,
        animation_duration: Optional[int] = None,
        divider_color: Optional[ColorValue] = None,
        indicator_color: Optional[ColorValue] = None,
        indicator_border_radius: Optional[BorderRadiusValue] = None,
        indicator_border_side: Optional[BorderSide] = None,
        indicator_padding: Optional[PaddingValue] = None,
        indicator_tab_size: Optional[bool] = None,
        is_secondary: Optional[bool] = None,
        label_color: Optional[ColorValue] = None,
        label_padding: Optional[PaddingValue] = None,
        label_text_style: Optional[TextStyle] = None,
        unselected_label_color: Optional[ColorValue] = None,
        unselected_label_text_style: Optional[TextStyle] = None,
        overlay_color: ControlStateValue[ColorValue] = None,
        divider_height: OptionalNumber = None,
        indicator_thickness: OptionalNumber = None,
        enable_feedback: Optional[str] = None,
        mouse_cursor: Optional[MouseCursor] = None,
        padding: Optional[PaddingValue] = None,
        splash_border_radius: Optional[BorderRadiusValue] = None,
        clip_behavior: Optional[ClipBehavior] = None,
        on_click: OptionalControlEventCallable = None,
        on_change: OptionalControlEventCallable = None,
        #
        # ConstrainedControl and AdaptiveControl
        #
        ref: Optional[Ref] = None,
        key: Optional[str] = None,
        width: OptionalNumber = None,
        height: OptionalNumber = None,
        left: OptionalNumber = None,
        top: OptionalNumber = None,
        right: OptionalNumber = None,
        bottom: OptionalNumber = None,
        expand: Union[None, bool, int] = None,
        expand_loose: Optional[bool] = None,
        col: Optional[ResponsiveNumber] = None,
        opacity: OptionalNumber = None,
        rotate: Optional[RotateValue] = None,
        scale: Optional[ScaleValue] = None,
        offset: Optional[OffsetValue] = None,
        aspect_ratio: OptionalNumber = None,
        animate_opacity: Optional[AnimationValue] = None,
        animate_size: Optional[AnimationValue] = None,
        animate_position: Optional[AnimationValue] = None,
        animate_rotation: Optional[AnimationValue] = None,
        animate_scale: Optional[AnimationValue] = None,
        animate_offset: Optional[AnimationValue] = None,
        on_animation_end: OptionalControlEventCallable = None,
        visible: Optional[bool] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
        adaptive: Optional[bool] = None,
    ):
        ConstrainedControl.__init__(
            self,
            ref=ref,
            key=key,
            width=width,
            height=height,
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            expand=expand,
            expand_loose=expand_loose,
            col=col,
            opacity=opacity,
            rotate=rotate,
            scale=scale,
            offset=offset,
            aspect_ratio=aspect_ratio,
            animate_opacity=animate_opacity,
            animate_size=animate_size,
            animate_position=animate_position,
            animate_rotation=animate_rotation,
            animate_scale=animate_scale,
            animate_offset=animate_offset,
            on_animation_end=on_animation_end,
            visible=visible,
            disabled=disabled,
            data=data,
        )

        AdaptiveControl.__init__(self, adaptive=adaptive)

        self.tabs = tabs
        self.selected_index = selected_index
        self.scrollable = scrollable
        self.tab_alignment = tab_alignment
        self.animation_duration = animation_duration
        self.divider_color = divider_color
        self.label_color = label_color
        self.unselected_label_color = unselected_label_color
        self.indicator_color = indicator_color
        self.indicator_border_radius = indicator_border_radius
        self.indicator_border_side = indicator_border_side
        self.indicator_padding = indicator_padding
        self.indicator_tab_size = indicator_tab_size
        self.overlay_color = overlay_color
        self.on_change = on_change
        self.divider_height = divider_height
        self.indicator_thickness = indicator_thickness
        self.enable_feedback = enable_feedback
        self.is_secondary = is_secondary
        self.mouse_cursor = mouse_cursor
        self.clip_behavior = clip_behavior
        self.label_padding = label_padding
        self.label_text_style = label_text_style
        self.unselected_label_text_style = unselected_label_text_style
        self.padding = padding
        self.splash_border_radius = splash_border_radius
        self.on_click = on_click

    def _get_control_name(self):
        return "tabs"

    def before_update(self):
        super().before_update()
        self._set_attr_json("overlayColor", self.__overlay_color, wrap_attr_dict=True)
        self._set_attr_json("indicatorBorderRadius", self.__indicator_border_radius)
        self._set_attr_json("indicatorBorderSide", self.__indicator_border_side)
        self._set_attr_json("indicatorPadding", self.__indicator_padding)
        self._set_attr_json("labelPadding", self.__label_padding)
        self._set_attr_json("labelTextStyle", self.__label_text_style)
        self._set_attr_json(
            "unselectedLabelTextStyle", self.__unselected_label_text_style
        )

    def _get_children(self):
        return self.__tabs

    def __contains__(self, item):
        return item in self.__tabs

    # tabs
    @property
    def tabs(self) -> List[Tab]:
        return self.__tabs

    @tabs.setter
    def tabs(self, value: Optional[List[Tab]]):
        self.__tabs = value if value is not None else []

    # on_change
    @property
    def on_change(self) -> OptionalControlEventCallable:
        return self._get_event_handler("change")

    @on_change.setter
    def on_change(self, handler: OptionalControlEventCallable):
        self._add_event_handler("change", handler)

    # selected_index
    @property
    def selected_index(self) -> int:
        return self._get_attr("selectedIndex", data_type="int", def_value=0)

    @selected_index.setter
    def selected_index(self, value: Optional[int]):
        self._set_attr("selectedIndex", value)

    # scrollable
    @property
    def scrollable(self) -> bool:
        return self._get_attr("scrollable", data_type="bool", def_value=True)

    @scrollable.setter
    def scrollable(self, value: Optional[bool]):
        self._set_attr("scrollable", value)

    # mouse_cursor
    @property
    def mouse_cursor(self) -> Optional[MouseCursor]:
        return self.__mouse_cursor

    @mouse_cursor.setter
    def mouse_cursor(self, value: Optional[MouseCursor]):
        self.__mouse_cursor = value
        self._set_enum_attr("mouseCursor", value, MouseCursor)

    # clip_behavior
    @property
    def clip_behavior(self) -> Optional[ClipBehavior]:
        return self.__clip_behavior

    @clip_behavior.setter
    def clip_behavior(self, value: Optional[ClipBehavior]):
        self.__clip_behavior = value
        self._set_enum_attr("clipBehavior", value, ClipBehavior)

    # is_secondary
    @property
    def is_secondary(self) -> bool:
        return self._get_attr("isSecondary", data_type="bool", def_value=False)

    @is_secondary.setter
    def is_secondary(self, value: Optional[bool]):
        self._set_attr("isSecondary", value)

    # tab_alignment
    @property
    def tab_alignment(self) -> Optional[TabAlignment]:
        return self.__tab_alignment

    @tab_alignment.setter
    def tab_alignment(self, value: Optional[TabAlignment]):
        self.__tab_alignment = value
        self._set_enum_attr("tabAlignment", value, TabAlignment)

    # animation_duration
    @property
    def animation_duration(self) -> Optional[int]:
        return self._get_attr("animationDuration", data_type="int")

    @animation_duration.setter
    def animation_duration(self, value: Optional[int]):
        self._set_attr("animationDuration", value)

    # divider_height
    @property
    def divider_height(self) -> float:
        return self._get_attr("dividerHeight", data_type="float", def_value=1.0)

    @divider_height.setter
    def divider_height(self, value: OptionalNumber):
        self._set_attr("dividerHeight", value)

    # enable_feedback
    @property
    def enable_feedback(self):
        return self._get_attr("enableFeedback")

    @enable_feedback.setter
    def enable_feedback(self, value: Optional[bool]):
        self._set_attr("enableFeedback", value)

    # indicator_thickness
    @property
    def indicator_thickness(self) -> float:
        return self._get_attr("indicatorThickness", data_type="float", def_value=2.0)

    @indicator_thickness.setter
    def indicator_thickness(self, value: OptionalNumber):
        assert value is None or value > 0, "indicator_thickness must be greater than 0"
        self._set_attr("indicatorThickness", value)

    # divider_color
    @property
    def divider_color(self) -> Optional[ColorValue]:
        return self.__divider_color

    @divider_color.setter
    def divider_color(self, value: Optional[ColorValue]):
        self.__divider_color = value
        self._set_enum_attr("dividerColor", value, ColorEnums)

    # indicator_color
    @property
    def indicator_color(self) -> Optional[ColorValue]:
        return self.__indicator_color

    @indicator_color.setter
    def indicator_color(self, value: Optional[ColorValue]):
        self.__indicator_color = value
        self._set_enum_attr("indicatorColor", value, ColorEnums)

    # indicator_border_radius
    @property
    def indicator_border_radius(self) -> Optional[BorderRadiusValue]:
        return self.__indicator_border_radius

    @indicator_border_radius.setter
    def indicator_border_radius(self, value: Optional[BorderRadiusValue]):
        self.__indicator_border_radius = value

    # indicator_border_side
    @property
    def indicator_border_side(self) -> Optional[BorderSide]:
        return self.__indicator_border_side

    @indicator_border_side.setter
    def indicator_border_side(self, value: Optional[BorderSide]):
        self.__indicator_border_side = value

    # indicator_padding
    @property
    def indicator_padding(self) -> Optional[PaddingValue]:
        return self.__indicator_padding

    @indicator_padding.setter
    def indicator_padding(self, value: Optional[PaddingValue]):
        self.__indicator_padding = value

    # indicator_tab_size
    @property
    def indicator_tab_size(self) -> bool:
        return self._get_attr("indicatorTabSize", data_type="bool", def_value=False)

    @indicator_tab_size.setter
    def indicator_tab_size(self, value: Optional[bool]):
        self._set_attr("indicatorTabSize", value)

    # label_color
    @property
    def label_color(self) -> Optional[ColorValue]:
        return self.__label_color

    @label_color.setter
    def label_color(self, value: Optional[ColorValue]):
        self.__label_color = value
        self._set_enum_attr("labelColor", value, ColorEnums)

    # unselected_label_color
    @property
    def unselected_label_color(self) -> Optional[ColorValue]:
        return self.__unselected_label_color

    @unselected_label_color.setter
    def unselected_label_color(self, value: Optional[ColorValue]):
        self.__unselected_label_color = value
        self._set_enum_attr("unselectedLabelColor", value, ColorEnums)

    # overlay_color
    @property
    def overlay_color(self) -> ControlStateValue[ColorValue]:
        return self.__overlay_color

    @overlay_color.setter
    def overlay_color(self, value: ControlStateValue[ColorValue]):
        self.__overlay_color = value

    # label_padding
    @property
    def label_padding(self) -> Optional[PaddingValue]:
        return self.__label_padding

    @label_padding.setter
    def label_padding(self, value: Optional[PaddingValue]):
        self.__label_padding = value

    # label_text_style
    @property
    def label_text_style(self) -> Optional[TextStyle]:
        return self.__label_text_style

    @label_text_style.setter
    def label_text_style(self, value: Optional[TextStyle]):
        self.__label_text_style = value

    # unselected_label_text_style
    @property
    def unselected_label_text_style(self) -> Optional[TextStyle]:
        return self.__unselected_label_text_style

    @unselected_label_text_style.setter
    def unselected_label_text_style(self, value: Optional[TextStyle]):
        self.__unselected_label_text_style = value

    # padding
    @property
    def padding(self) -> Optional[PaddingValue]:
        return self.__padding

    @padding.setter
    def padding(self, value: Optional[PaddingValue]):
        self.__padding = value

    # splash_border_radius
    @property
    def splash_border_radius(self) -> Optional[BorderRadiusValue]:
        return self.__splash_border_radius

    @splash_border_radius.setter
    def splash_border_radius(self, value: Optional[BorderRadiusValue]):
        self.__splash_border_radius = value

    # on_click
    @property
    def on_click(self) -> OptionalControlEventCallable:
        return self._get_event_handler("click")

    @on_click.setter
    def on_click(self, handler: OptionalControlEventCallable):
        self._add_event_handler("click", handler)



import asyncio
import logging
import threading
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Awaitable, Callable, Dict, Iterable, Optional, Union

import flet.core
from flet.core.locks import NopeLock
from flet.utils import is_pyodide

logger = logging.getLogger(flet.__name__)


class PubSubHub:
    def __init__(
        self,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        executor: Optional[ThreadPoolExecutor] = None,
    ):
        logger.debug("Creating new PubSubHub instance")
        self.__loop = loop
        self.__executor = executor
        self.__lock = threading.Lock() if not is_pyodide() else NopeLock()
        self.__subscribers: Dict[
            str, set[Union[Callable, Callable[..., Awaitable[Any]]]]
        ] = {}  # key: session_id, value: handler
        self.__topic_subscribers: Dict[
            str, Dict[str, set[Union[Callable, Callable[..., Awaitable[Any]]]]]
        ] = {}  # key: topic, value: dict[session_id, handler]
        self.__subscriber_topics: Dict[
            str, Dict[str, set[Union[Callable, Callable[..., Awaitable[Any]]]]]
        ] = {}  # key: session_id, value: dict[topic, handler]

    def send_all(self, message: Any):
        logger.debug(f"pubsub.send_all({message})")
        with self.__lock:
            for handlers in self.__subscribers.values():
                for handler in handlers:
                    self.__send(handler, [message])

    def send_all_on_topic(self, topic: str, message: Any):
        logger.debug(f"pubsub.send_all_on_topic({topic}, {message})")
        with self.__lock:
            if topic in self.__topic_subscribers:
                for handlers in self.__topic_subscribers[topic].values():
                    for handler in handlers:
                        self.__send(handler, [topic, message])

    def send_others(self, except_session_id: str, message: Any):
        logger.debug(f"pubsub.send_others({except_session_id}, {message})")
        with self.__lock:
            for session_id, handlers in self.__subscribers.items():
                if except_session_id != session_id:
                    for handler in handlers:
                        self.__send(handler, [message])

    def send_others_on_topic(self, except_session_id: str, topic: str, message: Any):
        logger.debug(
            f"pubsub.send_others_on_topic({except_session_id}, {topic}, {message})"
        )
        with self.__lock:
            if topic in self.__topic_subscribers:
                for session_id, handlers in self.__topic_subscribers[topic].items():
                    if except_session_id != session_id:
                        for handler in handlers:
                            self.__send(handler, [topic, message])

    def subscribe(self, session_id: str, handler: Callable):
        logger.debug(f"pubsub.subscribe({session_id})")
        with self.__lock:
            handlers = self.__subscribers.get(session_id)
            if handlers is None:
                handlers = set()
                self.__subscribers[session_id] = handlers
            handlers.add(handler)

    def subscribe_topic(
        self,
        session_id: str,
        topic: str,
        handler: Union[Callable, Callable[..., Awaitable[Any]]],
    ):
        logger.debug(f"pubsub.subscribe_topic({session_id}, {topic})")
        with self.__lock:
            self.__subscribe_topic(session_id, topic, handler)

    def __subscribe_topic(
        self,
        session_id: str,
        topic: str,
        handler: Union[Callable, Callable[..., Awaitable[Any]]],
    ):
        topic_subscribers = self.__topic_subscribers.get(topic)
        if topic_subscribers is None:
            topic_subscribers = {}
            self.__topic_subscribers[topic] = topic_subscribers
        handlers = topic_subscribers.get(session_id)
        if handlers is None:
            handlers = set()
            topic_subscribers[session_id] = handlers
        handlers.add(handler)
        subscriber_topics = self.__subscriber_topics.get(session_id)
        if subscriber_topics is None:
            subscriber_topics = {}
            self.__subscriber_topics[session_id] = subscriber_topics
        handlers = subscriber_topics.get(topic)
        if handlers is None:
            handlers = set()
            subscriber_topics[topic] = handlers
        handlers.add(handler)

    def unsubscribe(self, session_id: str):
        logger.debug(f"pubsub.unsubscribe({session_id})")
        with self.__lock:
            self.__unsubscribe(session_id)

    def unsubscribe_topic(self, session_id: str, topic: str):
        logger.debug(f"pubsub.unsubscribe({session_id}, {topic})")
        with self.__lock:
            self.__unsubscribe_topic(session_id, topic)

    def unsubscribe_all(self, session_id: str):
        logger.debug(f"pubsub.unsubscribe_all({session_id})")
        with self.__lock:
            self.__unsubscribe(session_id)
            if session_id in self.__subscriber_topics:
                for topic in list(self.__subscriber_topics[session_id].keys()):
                    self.__unsubscribe_topic(session_id, topic)

    def __unsubscribe(self, session_id: str):
        logger.debug(f"pubsub.__unsubscribe({session_id})")
        self.__subscribers.pop(session_id, None)

    def __unsubscribe_topic(self, session_id: str, topic: str):
        logger.debug(f"pubsub.__unsubscribe_topic({session_id}, {topic})")
        topic_subscribers = self.__topic_subscribers.get(topic)
        if topic_subscribers is not None:
            topic_subscribers.pop(session_id, None)
            if len(topic_subscribers) == 0:
                self.__topic_subscribers.pop(topic, None)
        subscriber_topics = self.__subscriber_topics.get(session_id)
        if subscriber_topics is not None:
            subscriber_topics.pop(topic, None)
            if len(subscriber_topics) == 0:
                self.__subscriber_topics.pop(session_id, None)

    def __send(
        self, handler: Union[Callable, Callable[..., Awaitable[Any]]], args: Iterable
    ):
        assert self.__loop, "PubSub event loop is not set"

        if asyncio.iscoroutinefunction(handler):
            asyncio.run_coroutine_threadsafe(handler(*args), self.__loop)
        else:
            if self.__executor:
                self.__loop.call_soon_threadsafe(
                    self.__loop.run_in_executor, self.__executor, handler, *args
                )
            else:
                handler(*args)



import json
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Callable, List, Optional

from flet.core.control import Control
from flet.core.control_event import ControlEvent
from flet.core.event_handler import EventHandler
from flet.core.ref import Ref
from flet.core.types import OptionalEventCallable


class FilePickerState(Enum):
    PICK_FILES = "pickFiles"
    SAVE_FILE = "saveFile"
    GET_DIRECTORY_PATH = "getDirectoryPath"


class FilePickerFileType(Enum):
    ANY = "any"
    MEDIA = "media"
    IMAGE = "image"
    VIDEO = "video"
    AUDIO = "audio"
    CUSTOM = "custom"


@dataclass
class FilePickerUploadFile:
    name: str
    upload_url: str
    id: int = None
    method: str = field(default="PUT")


@dataclass
class FilePickerFile:
    name: str
    path: str
    size: int
    id: int


class FilePickerResultEvent(ControlEvent):
    def __init__(self, e: ControlEvent):
        super().__init__(e.target, e.name, e.data, e.control, e.page)
        d = json.loads(e.data)
        self.path: Optional[str] = d.get("path")
        self.files: Optional[List[FilePickerFile]] = None
        files = d.get("files")
        if files is not None and isinstance(files, List):
            self.files = []
            for fd in files:
                self.files.append(FilePickerFile(**fd))


class FilePickerUploadEvent(ControlEvent):
    def __init__(self, e: ControlEvent):
        super().__init__(e.target, e.name, e.data, e.control, e.page)
        d = json.loads(e.data)
        self.file_name: str = d.get("file_name")
        self.progress: Optional[float] = d.get("progress")
        self.error: Optional[str] = d.get("error")


class FilePicker(Control):
    """
    A control that allows you to use the native file explorer to pick single or multiple files, with extensions filtering support and upload.

    Example:
    ```
    import flet as ft

    def main(page: ft.Page):
        def pick_files_result(e: ft.FilePickerResultEvent):
            selected_files.value = (
                ", ".join(map(lambda f: f.name, e.files)) if e.files else "Cancelled!"
            )
            selected_files.update()

        pick_files_dialog = ft.FilePicker(on_result=pick_files_result)
        selected_files = ft.Text()

        page.overlay.append(pick_files_dialog)

        page.add(
            ft.Row(
                [
                    ft.ElevatedButton(
                        "Pick files",
                        icon=ft.icons.UPLOAD_FILE,
                        on_click=lambda _: pick_files_dialog.pick_files(
                            allow_multiple=True
                        ),
                    ),
                    selected_files,
                ]
            )
        )

    ft.app(target=main)
    ```

    -----

    Online docs: https://flet.dev/docs/controls/filepicker
    """

    def __init__(
        self,
        on_result: Optional[Callable[[FilePickerResultEvent], None]] = None,
        on_upload: Optional[Callable[[FilePickerUploadEvent], None]] = None,
        #
        # Control
        #
        ref: Optional[Ref] = None,
        disabled: Optional[bool] = None,
        data: Any = None,
    ):
        Control.__init__(
            self,
            ref=ref,
            disabled=disabled,
            data=data,
        )

        def convert_result_event_data(e):
            self.__result = FilePickerResultEvent(e)
            return self.__result

        self.__on_result = EventHandler(convert_result_event_data)
        self._add_event_handler("result", self.__on_result.get_handler())

        self.__on_upload = EventHandler(lambda e: FilePickerUploadEvent(e))
        self._add_event_handler("upload", self.__on_upload.get_handler())

        self.__result: Optional[FilePickerResultEvent] = None
        self.__upload: List[FilePickerUploadFile] = []
        self.__allowed_extensions: Optional[List[str]] = None
        self.__state = None
        self.__file_type = None
        self.on_result = on_result
        self.on_upload = on_upload

    def _get_control_name(self):
        return "filepicker"

    def before_update(self):
        super().before_update()
        self._set_attr_json("allowedExtensions", self.__allowed_extensions)
        self._set_attr_json("upload", self.__upload)

    def pick_files(
        self,
        dialog_title: Optional[str] = None,
        initial_directory: Optional[str] = None,
        file_type: FilePickerFileType = FilePickerFileType.ANY,
        allowed_extensions: Optional[List[str]] = None,
        allow_multiple: Optional[bool] = False,
    ):
        self.state = FilePickerState.PICK_FILES
        self.dialog_title = dialog_title
        self.initial_directory = initial_directory
        self.file_type = file_type
        self.allowed_extensions = allowed_extensions
        self.allow_multiple = allow_multiple
        self.update()

    def save_file(
        self,
        dialog_title: Optional[str] = None,
        file_name: Optional[str] = None,
        initial_directory: Optional[str] = None,
        file_type: FilePickerFileType = FilePickerFileType.ANY,
        allowed_extensions: Optional[List[str]] = None,
    ):
        self.state = FilePickerState.SAVE_FILE
        self.dialog_title = dialog_title
        self.file_name = file_name
        self.initial_directory = initial_directory
        self.file_type = file_type
        self.allowed_extensions = allowed_extensions
        self.update()

    def get_directory_path(
        self,
        dialog_title: Optional[str] = None,
        initial_directory: Optional[str] = None,
    ):
        self.state = FilePickerState.GET_DIRECTORY_PATH
        self.dialog_title = dialog_title
        self.initial_directory = initial_directory
        self.update()

    def upload(self, files: List[FilePickerUploadFile]):
        self.__upload = files
        self.update()

    # state
    @property
    def state(self) -> Optional[FilePickerState]:
        return self.__state

    @state.setter
    def state(self, value: Optional[FilePickerState]):
        self.__state = value
        self._set_enum_attr("state", value, FilePickerState)

    # result
    @property
    def result(self) -> Optional[FilePickerResultEvent]:
        return self.__result

    # dialog_title
    @property
    def dialog_title(self) -> Optional[str]:
        return self._get_attr("dialogTitle")

    @dialog_title.setter
    def dialog_title(self, value: Optional[str]):
        self._set_attr("dialogTitle", value)

    # file_name
    @property
    def file_name(self) -> Optional[str]:
        return self._get_attr("fileName")

    @file_name.setter
    def file_name(self, value: Optional[str]):
        self._set_attr("fileName", value)

    # initial_directory
    @property
    def initial_directory(self):
        return self._get_attr("initialDirectory")

    @initial_directory.setter
    def initial_directory(self, value: Optional[str]):
        self._set_attr("initialDirectory", value)

    # file_type
    @property
    def file_type(self) -> FilePickerFileType:
        return self.__file_type

    @file_type.setter
    def file_type(self, value: FilePickerFileType):
        self.__file_type = value
        self._set_enum_attr("fileType", value, FilePickerFileType)

    # allowed_extensions
    @property
    def allowed_extensions(self) -> Optional[List[str]]:
        return self.__allowed_extensions

    @allowed_extensions.setter
    def allowed_extensions(self, value: Optional[List[str]]):
        self.__allowed_extensions = value

    # allow_multiple
    @property
    def allow_multiple(self) -> bool:
        return self._get_attr("allowMultiple", data_type="bool", def_value=False)

    @allow_multiple.setter
    def allow_multiple(self, value: Optional[bool]):
        self._set_attr("allowMultiple", value)

    # on_result
    @property
    def on_result(self) -> OptionalEventCallable[FilePickerResultEvent]:
        return self.__on_result.handler

    @on_result.setter
    def on_result(self, handler: OptionalEventCallable[FilePickerResultEvent]):
        self.__on_result.handler = handler

    # on_upload
    @property
    def on_upload(self):
        return self.__on_upload.handler

    @on_upload.setter
    def on_upload(self, handler: OptionalEventCallable[FilePickerUploadEvent]):
        self.__on_upload.handler = handler
